# Copyright 2024 NVIDIA CORPORATION & AFFILIATES
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0

import torch
# 4 block
import triton
import triton.language as tl
from triton.language.extra.cuda import libdevice

from ._division_transpose import fp8_division_transpose
from .common import (FP8_MAX_VALUE, SCALE_MIN_THRES, convert_str_to_fp8,
                     get_configs_io_block)

"""Per Tensor Quantize and Transpose Operator"""
"""Input uses floating point tensor"""
"""Output uses per-tensor quantization, returns a non-transpose version and a transpose version"""
"""The input can be 2D or 3D, but the calculation is performed in 2D"""


@triton.autotune(
    configs=[] + get_configs_io_block(),
    key=[
        "M",
        "N",
    ],
)
@triton.heuristics(
    {
        "BLOCK_SN": lambda args: args["BLOCK_N"] // args["QB"],
    }
)
@triton.jit
def _fp8_quantize_pertensor_transpose_kernel(
    output_scale_ptr,  # output
    input_ptr,  # input
    M,
    N,
    SN,
    QB: tl.constexpr,
    fp8_max,  # shape
    input_stride_0,
    input_stride_1,  # input stride
    s_output_stride_0,
    s_output_stride_1,  # scale of output stride
    SCALE_MIN_THRES: tl.constexpr,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
    BLOCK_SN: tl.constexpr,
):  # CUDA block size

    # Block PID
    pid = tl.program_id(0)
    NUM_BLOCK_N = tl.cdiv(N, BLOCK_N)
    pid_dim0 = pid // NUM_BLOCK_N
    pid_dim1 = pid % NUM_BLOCK_N

    # pointers
    input_block_ptr = tl.make_block_ptr(
        base=input_ptr,
        shape=(M, N),
        strides=(input_stride_0, input_stride_1),
        offsets=(pid_dim0 * BLOCK_M, pid_dim1 * BLOCK_N),
        block_shape=(BLOCK_M, BLOCK_N),
        order=(1, 0),
    )

    input = tl.load(input_block_ptr)
    input = input.to(tl.float32)

    output = tl.reshape(input, (BLOCK_M, BLOCK_SN, QB))

    # Quantize Scale calculation
    abs_output = tl.abs(output)
    max_val = tl.max(abs_output, axis=2) + SCALE_MIN_THRES
    scale_output = max_val / fp8_max
    scale_output = tl.reshape(scale_output, (BLOCK_M, BLOCK_SN, 1))

    scale_output = scale_output.to(output_scale_ptr.type.element_ty)
    scale_output = tl.reshape(scale_output, (BLOCK_M, BLOCK_SN))

    scale_output_ptr = tl.make_block_ptr(
        base=output_scale_ptr,
        shape=(M, SN),
        strides=(s_output_stride_0, s_output_stride_1),
        offsets=(pid_dim0 * BLOCK_M, pid_dim1 * BLOCK_SN),
        block_shape=(BLOCK_M, BLOCK_SN),
        order=(1, 0),
    )

    tl.store(scale_output_ptr, scale_output)


def fp8_quantize_pertensor_transpose(x, QB, fp8type, transpose_output_2d=False, stochastic=False, scale_dtype=torch.bfloat16):
    # Change batched 3D input to 2D
    batched = False
    if len(x.shape) == 3:
        batched = True
        BS = x.shape[0]
        x = x.reshape(-1, x.shape[-1])

    # defining the input and output tensor
    M, N = x.shape
    SN = N // QB

    fp8type = convert_str_to_fp8[fp8type]
    s_y = torch.empty((M, SN), dtype=scale_dtype, device=x.device)
    fp8MaxValue = FP8_MAX_VALUE[fp8type]  # E4M3 and E5M2 have different max value

    grid = lambda META: (triton.cdiv(M, META["BLOCK_M"]) * triton.cdiv(N, META["BLOCK_N"]),)

    _fp8_quantize_pertensor_transpose_kernel[grid](
        s_y,
        x,
        M,
        N,
        SN,
        QB,
        fp8MaxValue,
        x.stride(0),
        x.stride(1),
        s_y.stride(0),
        s_y.stride(1),
        SCALE_MIN_THRES=SCALE_MIN_THRES,
    )

    s_y_max = s_y.max()
    qy, s_y_max, qy_t = fp8_division_transpose(
        x, QB, fp8type, s_y_max, stochastic=stochastic
    )  # Stochastic Rounding happens here

    # Recover 2D to 3D
    if batched:
        qy = qy.reshape(BS, -1, qy.shape[-1])
        if not transpose_output_2d:
            qy_t = qy_t.reshape(BS, -1, qy_t.shape[-1])

    return qy, s_y_max, qy_t  # y_t is expected to be 2D tensor



# Copyright 2024 NVIDIA CORPORATION & AFFILIATES
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0

"""
Adapted from
[MosaiclML](https://github.com/mosaicml/examples.git) and
[minGPT](https://github.com/karpathy/minGPT.git)
"""

from __future__ import annotations

import logging
import math
import sys
from abc import abstractmethod
from collections import defaultdict
from functools import partial
from typing import (Callable, Dict, Iterable, List, NamedTuple, Optional,
                    Sequence, Set, Tuple, cast)

import torch
import torch.backends.cuda
import torch.nn as nn
import torch.nn.functional as F
from olmo.aliases import PathOrStr
from olmo.beam_search import (BeamSearch, Constraint, FinalSequenceScorer,
                              Sampler)
from olmo.config import (ActivationCheckpointingStrategy, ActivationType,
                         BlockType, CheckpointType, FSDPWrapStrategy,
                         InitFnType, LayerNormType, ModelConfig,
                         QuantActivationConfig, ShardedCheckpointerType,
                         TrainConfig)
from olmo.exceptions import OLMoConfigurationError
from olmo.initialization import init_normal
from olmo.model import (Activation, BufferCache, Dropout, LayerNorm,
                        LayerNormBase, OLMo, OLMoBlock, OLMoBlockGroup,
                        OLMoGenerateOutput, OLMoOutput, RMSLayerNorm,
                        RotaryEmbedding, _non_meta_init_device,
                        activation_checkpoint_function, alibi_attention_bias,
                        causal_attention_bias, get_causal_attention_bias,
                        should_checkpoint_block)
from olmo.torch_util import ensure_finite_, get_cumulative_document_lengths
from torch import einsum

from ..activation.real_quantization import (Coat_quantize_bgn, Coat_quantize_end,
                                 fp8_add_Ifp_Ifp_Ofp_Og16,
                                 fp8_add_Ifp_Ifp_Ofp_Opt, fp8_division,
                                 fp8_division_transpose, fp8_gelu_backward,
                                 fp8_gelu_forward,
                                 fp8_layernorm_noparam_backward,
                                 fp8_layernorm_noparam_forward,
                                 fp8_linear_backward, fp8_linear_forward,
                                 fp8_mul_backward, fp8_mul_forward,
                                 fp8_quantize, fp8_quantize_pertensor,
                                 fp8_quantize_pertensor_transpose,
                                 fp8_rmsnorm_backward, fp8_rmsnorm_forward,
                                 fp8_silu_backward, fp8_silu_forward,
                                 fp8_transpose)
from ..utils._fp8manager import FP8Manager
from ..utils._fp8_weightcache import FP8CacheWeightModule

if sys.version_info.minor > 8:
    from collections.abc import MutableMapping
elif sys.version_info.minor == 8:
    from typing import MutableMapping
else:
    raise SystemExit("This script supports Python 3.8 or higher")

__all__ = [
    "CoatOLMoBeforeAttentionResidual",
    "CoatOLMoAfterAttentionResidual",
    "CoatOLMoMLPResidual",
    "CoatOLMoBlock",
    "CoatOLMo",
]


log = logging.getLogger(__name__)


class CoatOLMoBeforeAttentionResidual(FP8CacheWeightModule):
    """
    This is a typical transformer attention module that contains (1) Residual (2) LayerNorm / RMSNorm (3) 1 * Linear layers
    """

    def __init__(self, config: ModelConfig, qargs: QuantActivationConfig, layer_id, fused_dims: tuple):
        super().__init__(config, qargs, layer_id)

        self.qargs = qargs
        self.fwobits = {
            "fabit": self.qargs.fabit,
            "fwbit": self.qargs.fwbit,
            "fobit": self.qargs.fobit,
            "babit": self.qargs.babit,
            "bwbit": self.qargs.bwbit,
            "bobit": self.qargs.bobit,
        }
        self.ln_normalized_shape = config.d_model
        self.att_proj = nn.Linear(config.d_model, sum(fused_dims), bias=config.include_bias, device=config.init_device)

        self.attn_norm = LayerNorm.build(config)

    def forward(self, re_x, x, s):
        if self.training:
            if self.qargs.weight_memory_efficient:
                # Prepare
                with torch.no_grad():
                    weight1_s = self.prepare_weight(self.att_proj.weight, "att_proj", FP8Manager.is_first_microbatch)
                return _CoatOLMoBeforeAttentionResidual.apply(
                    re_x,
                    x,
                    s,
                    self.att_proj.weight,
                    None,
                    None,
                    weight1_s,
                    self.qargs.group_size,
                    self.fwobits,
                    self.layer_id,
                    self.config,
                    self.qargs,
                )
            else:
                # Prepare
                with torch.no_grad():
                    weight1, weight1_t, weight1_s = self.prepare_weight(
                        self.att_proj.weight, "att_proj", FP8Manager.is_first_microbatch
                    )
                return _CoatOLMoBeforeAttentionResidual.apply(
                    re_x,
                    x,
                    s,
                    self.att_proj.weight,
                    weight1,
                    weight1_t,
                    weight1_s,
                    self.qargs.group_size,
                    self.fwobits,
                    self.layer_id,
                    self.config,
                    self.qargs,
                )
        else:
            return re_x, self.att_proj(self.attn_norm(re_x))


class _CoatOLMoBeforeAttentionResidual(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx,
        re_x,
        in_x,
        in_s,
        weight1_origin,
        weight1,
        weight1_t,
        weight1_s,
        group_size,
        fwobits,
        layer_id,
        config,
        qargs,
        eps=1e-5,
    ):
        # for autograd
        if fwobits["fabit"] == "E4M3":
            # in_x = in_x.to(torch.float8_e4m3fn)
            in_x = in_x.view(torch.float8_e4m3fn)
        else:
            raise ValueError("fabit should be E4M3")

        # LayerNorm
        ln_x, ln_s, ln_x_t, ln_utils = fp8_layernorm_noparam_forward(
            in_x, in_s, group_size, eps, transpose_output_2d=True
        )

        # Linear Layer QKV Projection
        if qargs.weight_memory_efficient:
            assert weight1 is None  # memory efficient
            weight1, weight1_s = fp8_division(weight1_origin, qargs.group_size, fwobits["fwbit"], weight1_s)
        fc1_x = fp8_linear_forward(ln_x, ln_s, weight1, weight1_s, False, group_size)

        # ==================== save for backward ====================
        ctx.save_for_backward(in_x, in_s, ln_x_t, ln_s)
        if qargs.weight_memory_efficient:
            assert weight1_t is None
            ctx.weight = weight1_origin, weight1_s
        else:
            ctx.weight = weight1_t, weight1_s
        ctx.group_size = group_size
        ctx.ln_utils = ln_utils
        ctx.utils = fwobits, layer_id, config, qargs

        return re_x, fc1_x

    @staticmethod
    def backward(ctx, fp_grad, flash_g):
        in_x, in_s, ln_x_t, ln_s = ctx.saved_tensors
        weight1_t, weight1_s = ctx.weight
        group_size = ctx.group_size
        mean, rstd, num_warps = ctx.ln_utils
        fwobits, layer_id, config, qargs = ctx.utils

        # ==================== Begin backward ====================
        # Quantize the RoPE and FlashAttention Output. grad_input and grad_weight requires different data layout.
        flash_g, flash_gs, flash_g_t = fp8_quantize_pertensor_transpose(
            flash_g, group_size, fwobits["babit"], transpose_output_2d=True, stochastic=False
        )

        # Linear Layer QKV Projection
        if qargs.weight_memory_efficient:
            weight1_t, weight1_s = fp8_division_transpose(
                weight1_t, qargs.group_size, fwobits["fwbit"], weight1_s, only_transposed=True
            )
        fc1_g, att_proj_wg = fp8_linear_backward(
            ln_x_t, ln_s, flash_g, flash_gs, flash_g_t, weight1_t, weight1_s, group_size
        )

        # LayerNorm
        in_g = fp8_layernorm_noparam_backward(in_x, in_s, fc1_g, group_size, mean, rstd, num_warps)

        # Add the gradient together, and prepare the input of the next layer.
        re_g, (in_g, in_sg, in_sg_g16) = fp8_add_Ifp_Ifp_Ofp_Opt(
            fp_grad, in_g, group_size, fwobits["babit"], stochastic=False
        )

        # for autograd. forward's data type should be the same of backward tensor. this will not change the actual binary representation.
        in_g = in_g.view(torch.float8_e4m3fn)

        # Although the next operator is a linear layer in MLPResidual module, we return in_sg_g16 to make the size compatible with the forward. Otherwise it will not pass autograd.
        return re_g, in_g, in_sg_g16, att_proj_wg, None, None, None, None, None, None, None, None, None


class CoatOLMoAfterAttentionResidual(FP8CacheWeightModule):
    """
    This is a typical transformer attention module that contains (1) Residual (2) 1 * Linear layers
    """

    def __init__(self, config: ModelConfig, qargs: QuantActivationConfig, layer_id):
        super().__init__(config, qargs, layer_id)

        self.qargs = qargs
        self.fwobits = {
            "fabit": self.qargs.fabit,
            "fwbit": self.qargs.fwbit,
            "fobit": self.qargs.fobit,
            "babit": self.qargs.babit,
            "bwbit": self.qargs.bwbit,
            "bobit": self.qargs.bobit,
        }
        self.attn_out = nn.Linear(config.d_model, config.d_model, bias=config.include_bias, device=config.init_device)

    def forward(self, re_x, in_x):
        if self.training:
            if self.qargs.weight_memory_efficient:
                # prepare for the weight
                with torch.no_grad():
                    weight2_s = self.prepare_weight(self.attn_out.weight, "attn_out", FP8Manager.is_first_microbatch)

                return _CoatOLMoAfterAttentionResidual.apply(
                    re_x,
                    in_x,
                    self.attn_out.weight,
                    None,
                    None,
                    weight2_s,
                    self.qargs.group_size,
                    self.fwobits,
                    self.layer_id,
                    self.config,
                    self.qargs,
                )
            else:
                # prepare for the weight
                with torch.no_grad():
                    weight2, weight2_t, weight2_s = self.prepare_weight(
                        self.attn_out.weight, "attn_out", FP8Manager.is_first_microbatch
                    )

                return _CoatOLMoAfterAttentionResidual.apply(
                    re_x,
                    in_x,
                    self.attn_out.weight,
                    weight2,
                    weight2_t,
                    weight2_s,
                    self.qargs.group_size,
                    self.fwobits,
                    self.layer_id,
                    self.config,
                    self.qargs,
                )
        else:
            return re_x + self.attn_out(in_x), None, None


class _CoatOLMoAfterAttentionResidual(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx, re_x, flash_x, weight2_origin, weight2, weight2_t, weight2_s, group_size, fwobits, layer_id, config, qargs
    ):
        # Quantize the FlashAttention Output
        flash_qx, flash_s, _ = fp8_quantize_pertensor(
            flash_x, group_size, fwobits["fabit"]
        )  # Modified to make it memory efficient

        # # Attention Projection Linear Layer
        if qargs.weight_memory_efficient:
            assert weight2 is None  # memory efficient
            weight2, weight2_s = fp8_division(weight2_origin, qargs.group_size, fwobits["fwbit"], weight2_s)
        fc2_x = fp8_linear_forward(flash_qx, flash_s, weight2, weight2_s, False, group_size)  #

        # import IPython
        # IPython.embed()
        # Add the activations together
        fp_x, (out_x, out_s) = fp8_add_Ifp_Ifp_Ofp_Og16(re_x, fc2_x, flash_qx.dtype, group_size)

        # ==================== save for backward ====================
        ctx.save_for_backward(flash_x, flash_s)
        if qargs.weight_memory_efficient:
            assert weight2_t is None
            ctx.weight = weight2_origin, weight2_s
        else:
            ctx.weight = weight2_t, weight2_s
        ctx.group_size = group_size
        ctx.fwobits = fwobits
        ctx.utils = fwobits, layer_id, config, qargs

        # For autograd
        out_x = out_x.view(torch.float8_e4m3fn)

        return fp_x, out_x, out_s

    @staticmethod
    def backward(ctx, fp_grad, out_g, out_gs):
        flash_x, flash_s = ctx.saved_tensors
        weight2_t, weight2_s = ctx.weight
        group_size = ctx.group_size
        fwobits = ctx.fwobits
        fwobits, layer_id, config, qargs = ctx.utils

        # for autograd
        if fwobits["babit"] == "E5M2":
            # out_g = out_g.to(torch.float8_e5m2)
            out_g = out_g.view(torch.float8_e5m2)
        else:
            raise ValueError("babit should be E5M2")
        out_gs_max = out_gs.max()

        # ==================== Begin backward ====================
        # Output Projection
        out_g_t = fp8_transpose(out_g, transpose_output_2d=True)

        # We do not save an extra flash_x to save the memory usage
        flash_x_t, flash_s = fp8_division_transpose(
            flash_x, group_size, fwobits["fabit"], flash_s, stochastic=False, only_transposed=True
        )

        if qargs.weight_memory_efficient:
            weight2_t, weight2_s = fp8_division_transpose(
                weight2_t, qargs.group_size, fwobits["fwbit"], weight2_s, only_transposed=True
            )
        fc2_g, attn_out_wg = fp8_linear_backward(
            flash_x_t, flash_s, out_g, out_gs_max, out_g_t, weight2_t, weight2_s, group_size
        )

        return fp_grad, fc2_g, attn_out_wg, None, None, None, None, None, None, None, None


class CoatOLMoMLPResidual(FP8CacheWeightModule):
    """
    This is a typical transformer attention module that contains (1) Residual (2) LayerNorm / RMSNorm (3) 2 / 3 * Linear layers
    (4) GELU / Silu Activation
    """

    def __init__(self, config: ModelConfig, qargs: QuantActivationConfig, layer_id, hidden_size: int):
        super().__init__(config, qargs, layer_id)

        self.qargs = qargs
        self.fwobits = {
            "fabit": self.qargs.fabit,
            "fwbit": self.qargs.fwbit,
            "fobit": self.qargs.fobit,
            "babit": self.qargs.babit,
            "bwbit": self.qargs.bwbit,
            "bobit": self.qargs.bobit,
        }
        self.ln_normalized_shape = config.d_model
        self.act_output_multiplier = 0.5 if config.activation_type == ActivationType.swiglu else 1
        self.ff_proj = nn.Linear(config.d_model, hidden_size, bias=config.include_bias, device=config.init_device)
        self.ff_out = nn.Linear(
            int(self.act_output_multiplier * hidden_size),
            config.d_model,
            bias=config.include_bias,
            device=config.init_device,
        )
        self.training = True

        # below is only used when training = False
        self.ff_norm = LayerNorm.build(config)
        self.act = Activation.build(config)
        assert (self.act.output_multiplier * hidden_size) % 1 == 0

    def forward(self, re_x, x, s):
        if self.training:
            if self.qargs.weight_memory_efficient:  # prepare for the weight
                with torch.no_grad():
                    weight1_s = self.prepare_weight(self.ff_proj.weight, "ff_proj", FP8Manager.is_first_microbatch)
                    weight2_s = self.prepare_weight(self.ff_out.weight, "ff_out", FP8Manager.is_first_microbatch)

                return _CoatOLMoMLPResidual.apply(
                    re_x,
                    x,
                    s,
                    self.ff_proj.weight,
                    None,
                    None,
                    weight1_s,
                    self.ff_out.weight,
                    None,
                    None,
                    weight2_s,
                    self.qargs.group_size,
                    self.fwobits,
                    self.layer_id,
                    self.config,
                    self.qargs,
                )
            else:
                # prepare for the weight
                with torch.no_grad():
                    weight1, weight1_t, weight1_s = self.prepare_weight(
                        self.ff_proj.weight, "ff_proj", FP8Manager.is_first_microbatch
                    )
                    weight2, weight2_t, weight2_s = self.prepare_weight(
                        self.ff_out.weight, "ff_out", FP8Manager.is_first_microbatch
                    )

                return _CoatOLMoMLPResidual.apply(
                    re_x,
                    x,
                    s,
                    self.ff_proj.weight,
                    weight1,
                    weight1_t,
                    weight1_s,
                    self.ff_out.weight,
                    weight2,
                    weight2_t,
                    weight2_s,
                    self.qargs.group_size,
                    self.fwobits,
                    self.layer_id,
                    self.config,
                    self.qargs,
                )
        else:
            og_x = re_x
            re_x = self.ff_norm(re_x)
            re_x = self.ff_proj(re_x)
            re_x = self.act(re_x)
            re_x = self.ff_out(re_x)
            re_x = og_x + re_x
            return re_x, None, None


class _CoatOLMoMLPResidual(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx,
        re_x,
        in_x,
        in_s,
        weight1_origin,
        weight1,
        weight1_t,
        weight1_s,
        weight2_origin,
        weight2,
        weight2_t,
        weight2_s,
        group_size,
        fwobits,
        layer_id,
        config,
        qargs,
        eps=1e-5,
    ):
        # For autograd
        if fwobits["fabit"] == "E4M3":
            # in_x = in_x.to(torch.float8_e4m3fn)
            in_x = in_x.view(torch.float8_e4m3fn)
        else:
            raise ValueError("fabit should be E4M3")

        # LayerNorm
        ln_x, ln_s, ln_x_t, ln_utils = fp8_layernorm_noparam_forward(
            in_x, in_s, group_size, eps, transpose_output_2d=True
        )

        # Linear Layer of Up Projection and Gate Projection. They are fused as one linear layer.
        if qargs.weight_memory_efficient:
            assert weight1 is None  # memory efficient
            weight1, weight1_s = fp8_division(weight1_origin, qargs.group_size, fwobits["fwbit"], weight1_s)
        fc1_x, fc1_s = fp8_linear_forward(ln_x, ln_s, weight1, weight1_s, True, group_size)

        # NOTE: Becareful of the order
        up_x, gate_x = fc1_x.chunk(2, dim=-1)
        up_s, gate_s = fc1_s.chunk(2, dim=-1)

        # silu Activation
        silu_x, silu_s = fp8_silu_forward(gate_x, gate_s, group_size)

        # Element-wise Multiplication
        mul_x, mul_s, mul_x_t = fp8_mul_forward(silu_x, silu_s, up_x, up_s, group_size, transpose_output_2d=True)

        # Output Projection
        if weight2 is None:  # memory efficient
            weight2, weight2_s = fp8_division(weight2_origin, qargs.group_size, fwobits["fwbit"], weight2_s)
        fc2_x = fp8_linear_forward(mul_x, mul_s, weight2, weight2_s, False, group_size)

        # Add the activation together
        fp_x, (out_x, out_s) = fp8_add_Ifp_Ifp_Ofp_Og16(re_x, fc2_x, mul_x.dtype, group_size)

        # ==================== save for backward ====================
        ctx.save_for_backward(in_x, in_s, ln_x_t, ln_s, gate_x, gate_s, up_x, up_s, silu_x, silu_s, mul_x_t, mul_s)

        ctx.weight = (weight1_t, weight1_s, weight2_t, weight2_s)
        if (
            qargs.weight_memory_efficient
        ):  # Weight_1/2_origin will not be saved twice, so it will be more memory efficient.
            assert weight1_t is None
            ctx.weight = (weight1_origin, weight1_s, weight2_origin, weight2_s)
        else:  # Weight1/2_t is different from the origin weight, so saving it will consumes additional memory footprint.
            ctx.weight = (weight1_t, weight1_s, weight2_t, weight2_s)

        ctx.group_size = group_size
        ctx.ln_utils = ln_utils
        ctx.utils = fwobits, layer_id, config, qargs

        out_x = out_x.view(torch.float8_e4m3fn)

        return fp_x, out_x, out_s

    @staticmethod
    def backward(ctx, fp_grad, out_g, out_gs):
        fwobits, layer_id, config, qargs = ctx.utils

        in_x, in_s, ln_x_t, ln_s, gate_x, gate_s, up_x, up_s, silu_x, silu_s, mul_x_t, mul_s = ctx.saved_tensors

        (weight1_t, weight1_s, weight2_t, weight2_s) = ctx.weight
        group_size = ctx.group_size
        mean, rstd, num_warps = ctx.ln_utils
        fwobits, layer_id, config, qargs = ctx.utils

        # For autograd
        if fwobits["babit"] == "E5M2":
            # out_g = out_g.to(torch.float8_e5m2)
            out_g = out_g.view(torch.float8_e5m2)
        else:
            raise ValueError("babit should be E5M2")
        out_gs_max = out_gs.max()

        # ==================== Begin backward ====================
        # Output Projection
        out_gs = out_gs.max()
        out_g_t = fp8_transpose(out_g, transpose_output_2d=True)

        if qargs.weight_memory_efficient:
            weight2_t, weight2_s = fp8_division_transpose(
                weight2_t, qargs.group_size, fwobits["fwbit"], weight2_s, only_transposed=True
            )
        fc2_g, weight2_grad = fp8_linear_backward(
            mul_x_t, mul_s, out_g, out_gs_max, out_g_t, weight2_t, weight2_s, group_size
        )

        # [MEM TEST]
        del out_g, out_g_t, weight2_t

        # Element-wise Multiplication, 1 means gate, 2 means up
        mul_g1, (mul_g2, mul_gs2) = fp8_mul_backward(silu_x, silu_s, up_x, up_s, fc2_g, group_size, fwobits["babit"])

        # Silu activation
        silu_g, silu_gs = fp8_silu_backward(gate_x, gate_s, mul_g1, group_size, fwobits["babit"])

        # Prepare the input of Linear Layer. NOTE: Becareful of the order
        gateup_g = torch.cat([mul_g2, silu_g], dim=-1)
        gateup_gs = torch.cat([mul_gs2, silu_gs])
        gateup_gs = torch.max(gateup_gs)

        gateup_g, gateup_gs, gateup_g_t = fp8_division_transpose(
            gateup_g, group_size, fwobits["babit"], gateup_gs, stochastic=False
        )

        # Linear Layer of Up and Gate Projection
        if qargs.weight_memory_efficient:
            weight1_t, weight1_s = fp8_division_transpose(
                weight1_t, group_size, fwobits["fwbit"], weight1_s, only_transposed=True
            )
        fc1_g, weight1_grad = fp8_linear_backward(
            ln_x_t, ln_s, gateup_g, gateup_gs, gateup_g_t, weight1_t, weight1_s, group_size
        )

        # layerNorm
        in_g = fp8_layernorm_noparam_backward(in_x, in_s, fc1_g, group_size, mean, rstd, num_warps)

        # Add the gradient together
        re_g, (in_g, in_sg, in_sg_g16) = fp8_add_Ifp_Ifp_Ofp_Opt(
            fp_grad, in_g, group_size, fwobits["babit"], stochastic=False
        )

        in_g = in_g.view(torch.float8_e4m3fn)

        return (
            re_g,
            in_g,
            in_sg_g16,
            weight1_grad,
            None,
            None,
            None,
            weight2_grad,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
        )


class CoatOLMoBlock(nn.Module):
    """
    A base class for transformer block implementations.
    """

    def __init__(self, layer_id: int, config: ModelConfig, qargs: QuantActivationConfig, cache: BufferCache):
        super().__init__()
        self.layer_id = layer_id
        self.config = config
        self.qargs = qargs
        self.hidden_size = (
            config.mlp_hidden_size if config.mlp_hidden_size is not None else config.mlp_ratio * config.d_model
        )
        self.__cache = cache
        assert config.d_model % config.n_heads == 0

        self._activation_checkpoint_fn: Callable | None = None

        # Dropout.
        self.dropout = Dropout(config.residual_dropout)

        # Layer norms.
        self.k_norm: LayerNormBase | None = None
        self.q_norm: LayerNormBase | None = None
        if config.attention_layer_norm:
            assert config.effective_n_kv_heads is not None
            self.k_norm = LayerNormBase.build(
                config,
                size=(config.d_model // config.n_heads) * config.effective_n_kv_heads,
                elementwise_affine=config.attention_layer_norm_with_affine,
            )
            self.q_norm = LayerNormBase.build(config, elementwise_affine=config.attention_layer_norm_with_affine)

        # Make sure QKV clip coefficient is positive, otherwise it's not well-defined.
        if config.clip_qkv is not None:
            assert config.clip_qkv > 0

        # Activation function.
        self.act = Activation.build(config)
        assert (self.act.output_multiplier * self.hidden_size) % 1 == 0

        if not self.qargs.use_quantize_model:
            # Attention output projection.
            self.attn_out = nn.Linear(
                config.d_model, config.d_model, bias=config.include_bias, device=config.init_device
            )

            # Feed-forward output projection.
            self.ff_out = nn.Linear(
                int(self.act.output_multiplier * self.hidden_size),
                config.d_model,
                bias=config.include_bias,
                device=config.init_device,
            )
            self.ff_out._is_residual = True  # type: ignore

        # Rotary embeddings.
        if self.config.rope:
            self.rotary_emb = RotaryEmbedding(config, self.__cache)

        self.flash_attn_func = None
        self.flash_attn_varlen_func = None
        if config.flash_attention:
            try:
                from flash_attn import (flash_attn_func,  # type: ignore
                                        flash_attn_varlen_func)

                self.flash_attn_func = flash_attn_func
                self.flash_attn_varlen_func = flash_attn_varlen_func
            except ModuleNotFoundError:
                pass

    def reset_parameters(self):
        if self.k_norm is not None:
            self.k_norm.reset_parameters()
        if self.q_norm is not None:
            self.q_norm.reset_parameters()

        if not self.qargs.use_quantize_model:
            if self.config.init_fn == InitFnType.normal:
                attn_out_std = ff_out_std = self.config.init_std
                cutoff_factor = self.config.init_cutoff_factor

            elif self.config.init_fn == InitFnType.mitchell:
                attn_out_std = 1 / (math.sqrt(2 * self.config.d_model * (self.layer_id + 1)))
                ff_out_std = 1 / (math.sqrt(2 * self.ff_out.in_features * (self.layer_id + 1)))
                cutoff_factor = self.config.init_cutoff_factor or 3.0

            elif self.config.init_fn == InitFnType.full_megatron:
                attn_out_std = ff_out_std = self.config.init_std / math.sqrt(2.0 * self.config.n_layers)
                cutoff_factor = self.config.init_cutoff_factor or 3.0

            else:
                raise NotImplementedError(self.config.init_fn)

            init_normal(self.attn_out, std=attn_out_std, init_cutoff_factor=cutoff_factor)
            init_normal(self.ff_out, std=ff_out_std, init_cutoff_factor=cutoff_factor)

    def set_activation_checkpointing(
        self, strategy: ActivationCheckpointingStrategy | None, checkpoint_func: Callable | None = None
    ):
        if strategy == ActivationCheckpointingStrategy.fine_grained:
            self._activation_checkpoint_fn = checkpoint_func or activation_checkpoint_function(self.config)
        else:
            self._activation_checkpoint_fn = None

    @classmethod
    def _cast_attn_bias(cls, bias: torch.Tensor, input_dtype: torch.dtype) -> torch.Tensor:
        target_dtype = input_dtype
        # NOTE: `is_autocast_enabled()` only checks for CUDA autocast, so we use the separate function
        # `is_autocast_cpu_enabled()` for CPU autocast.
        # See https://github.com/pytorch/pytorch/issues/110966.
        if bias.device.type == "cuda" and torch.is_autocast_enabled():
            target_dtype = torch.get_autocast_gpu_dtype()
        elif bias.device.type == "cpu" and torch.is_autocast_cpu_enabled():
            target_dtype = torch.get_autocast_cpu_dtype()
        if bias.dtype != target_dtype:
            bias = bias.to(target_dtype)
            ensure_finite_(bias, check_neg_inf=True, check_pos_inf=False)
        return bias

    def _scaled_dot_product_attention(
        self,
        q: torch.Tensor,
        k: torch.Tensor,
        v: torch.Tensor,
        attn_mask: torch.Tensor | None = None,
        dropout_p: float = 0.0,
        is_causal: bool = False,
        max_doc_len: int | None = None,
        cu_doc_lens: torch.Tensor | None = None,
    ) -> torch.Tensor:
        """
        Computes scaled dot product attention on query, key and value tensors, using an optional
        attention mask if passed, and applying dropout if a probability greater than 0.0 is specified.
        """
        if max_doc_len is not None and cu_doc_lens is not None:
            assert self.flash_attn_varlen_func is not None, "flash-attn is required for document masking"
            assert attn_mask is None, "attn-mask is currently not supported with document masking"
            B, T, D = q.size(0), q.size(2), q.size(3)
            r = self.flash_attn_varlen_func(
                q.transpose(1, 2).view(B * T, -1, D),
                k.transpose(1, 2).view(B * T, -1, D),
                v.transpose(1, 2).view(B * T, -1, D),
                cu_doc_lens,
                cu_doc_lens,
                max_doc_len,
                max_doc_len,
                dropout_p=dropout_p,
                causal=is_causal,
            )
            return r.view(B, T, -1, D).transpose(1, 2)
        elif self.flash_attn_func is not None and attn_mask is None:
            r = self.flash_attn_func(
                q.transpose(1, 2), k.transpose(1, 2), v.transpose(1, 2), dropout_p=dropout_p, causal=is_causal
            )
            return r.transpose(1, 2)
        else:
            # torch's sdpa doesn't support GQA, so we're doing this
            assert k.size(1) == v.size(1)
            num_kv_heads = k.size(1)
            num_q_heads = q.size(1)
            if num_q_heads != num_kv_heads:
                assert num_q_heads % num_kv_heads == 0
                k = k.repeat_interleave(num_q_heads // num_kv_heads, dim=1, output_size=num_q_heads)
                v = v.repeat_interleave(num_q_heads // num_kv_heads, dim=1, output_size=num_q_heads)

            return F.scaled_dot_product_attention(
                q,
                k,
                v,
                attn_mask=attn_mask,
                dropout_p=dropout_p,
                is_causal=is_causal,
            )

    def attention(
        self,
        q: torch.Tensor,
        k: torch.Tensor,
        v: torch.Tensor,
        attention_bias: torch.Tensor | None = None,
        layer_past: tuple[torch.Tensor, torch.Tensor] | None = None,
        use_cache: bool = False,
        max_doc_len: int | None = None,
        cu_doc_lens: torch.Tensor | None = None,
    ) -> tuple[torch.Tensor, tuple[torch.Tensor, torch.Tensor] | None]:
        B, T, C = q.size()  # batch size, sequence length, d_model
        dtype = k.dtype

        # Optionally apply layer norm to keys and queries.
        if self.q_norm is not None and self.k_norm is not None:
            q = self.q_norm(q).to(dtype=dtype)
            k = self.k_norm(k).to(dtype=dtype)

        # Move head forward to be next to the batch dim.
        # shape: (B, nh, T, hs)
        q = q.view(B, T, self.config.n_heads, C // self.config.n_heads).transpose(1, 2)
        # shape: (B, n_kv_h, T, hs)
        k = k.view(B, T, self.config.effective_n_kv_heads, C // self.config.n_heads).transpose(1, 2)
        # shape: (B, n_kv_h, T, hs)
        v = v.view(B, T, self.config.effective_n_kv_heads, C // self.config.n_heads).transpose(1, 2)

        if layer_past is not None:
            past_key, past_value = layer_past
            k = torch.cat((past_key, k), dim=-2)
            v = torch.cat((past_value, v), dim=-2)

        present = (k, v) if use_cache else None
        query_len, key_len = q.shape[-2], k.shape[-2]  # could be different if layer_past not None

        if self.config.rope:
            # Apply rotary embeddings.
            q, k = self.rotary_emb(q, k)

        if attention_bias is not None:
            # Resize and cast attention bias.
            # The current dtype of the attention bias might not match the dtype that the SDP attn function will
            # run in if AMP is enabled, and this can be a problem if some tokens are masked out due to padding
            # as down-casting the attention bias to the autocast precision will result in -infs, which will
            # cause the SDP attn function to produce NaNs.
            attention_bias = self._cast_attn_bias(attention_bias[:, :, key_len - query_len : key_len, :key_len], dtype)

        # Get the attention scores.
        # shape: (B, nh, T, hs)
        att = self._scaled_dot_product_attention(
            q,
            k,
            v,
            attn_mask=attention_bias,
            dropout_p=0.0 if not self.training else self.config.attention_dropout,
            is_causal=attention_bias is None,
            max_doc_len=max_doc_len,
            cu_doc_lens=cu_doc_lens,
        )

        # Re-assemble all head outputs side-by-side.
        att = att.transpose(1, 2).contiguous().view(B, T, C)

        # Apply output projection. NOTE: We move the attn output outside of this attention function
        return att, present

    @abstractmethod
    def forward(
        self,
        x: torch.Tensor,
        attention_bias: torch.FloatTensor | None = None,
        layer_past: tuple[torch.Tensor, torch.Tensor] | None = None,
        use_cache: bool = False,
        max_doc_len: int | None = None,
        cu_doc_lens: torch.Tensor | None = None,
    ) -> tuple[torch.Tensor, tuple[torch.Tensor, torch.Tensor] | None]:
        raise NotImplementedError

    @classmethod
    def build(cls, layer_id: int, config: ModelConfig, qargs: QuantActivationConfig, cache: BufferCache) -> OLMoBlock:
        if config.block_type == BlockType.sequential:
            return CoatOLMoSequentialBlock(layer_id, config, qargs, cache)
        else:
            raise NotImplementedError(f"Unknown block type: '{config.block_type}'")


class CoatOLMoSequentialBlock(CoatOLMoBlock):
    """
    This is a typical transformer block where the output is computed as ``MLP(LN(x + Attention(LN(x))))``
    (plus another skip connection). To compute it as ``LN(MLP(x + LN(Attention(x))))``,
    use the flag `norm_after`.
    """

    def __init__(self, layer_id: int, config: ModelConfig, qargs: QuantActivationConfig, cache: BufferCache):
        super().__init__(layer_id, config, qargs, cache)
        # Attention input projection. Projects x -> (q, k, v)

        assert not self.config.norm_after, "COAT currently does not support PostNorm"

        head_dim = config.d_model // config.n_heads
        self.fused_dims = (
            config.d_model,
            config.effective_n_kv_heads * head_dim,
            config.effective_n_kv_heads * head_dim,
        )

        if self.qargs.use_quantize_model:
            self.BeforeAttention = CoatOLMoBeforeAttentionResidual(config, qargs, self.layer_id, self.fused_dims)
            self.AfterAttention = CoatOLMoAfterAttentionResidual(config, qargs, self.layer_id)
            self.MLPResidual = CoatOLMoMLPResidual(config, qargs, self.layer_id, self.hidden_size)
        else:
            self.att_proj = nn.Linear(
                config.d_model, sum(self.fused_dims), bias=config.include_bias, device=config.init_device
            )
            # Feed-forward input projection.
            self.ff_proj = nn.Linear(
                config.d_model, self.hidden_size, bias=config.include_bias, device=config.init_device
            )

        # Layer norms.
        self.attn_norm = LayerNorm.build(config, size=config.d_model)
        self.ff_norm = LayerNorm.build(config, size=config.d_model)

    def reset_parameters(self):
        super().reset_parameters()
        self.attn_norm.reset_parameters()
        self.ff_norm.reset_parameters()
        # NOTE: the standard deviation for these weights does not depend on the layer.

        if self.qargs.use_quantize_model:  # The initialization appears here, not in CoatOLMoBlock's reset_parameters
            if self.config.init_fn == InitFnType.normal:
                attn_out_std = ff_out_std = self.config.init_std
                cutoff_factor = self.config.init_cutoff_factor

            elif self.config.init_fn == InitFnType.mitchell:
                attn_out_std = 1 / (math.sqrt(2 * self.config.d_model * (self.layer_id + 1)))
                ff_out_std = 1 / (math.sqrt(2 * self.MLPResidual.ff_out.in_features * (self.layer_id + 1)))
                cutoff_factor = self.config.init_cutoff_factor or 3.0

            elif self.config.init_fn == InitFnType.full_megatron:
                attn_out_std = ff_out_std = self.config.init_std / math.sqrt(2.0 * self.config.n_layers)
                cutoff_factor = self.config.init_cutoff_factor or 3.0

            else:
                raise NotImplementedError(self.config.init_fn)

            init_normal(self.AfterAttention.attn_out, std=attn_out_std, init_cutoff_factor=cutoff_factor)
            init_normal(self.MLPResidual.ff_out, std=ff_out_std, init_cutoff_factor=cutoff_factor)

        if self.config.init_fn == InitFnType.normal:
            std = self.config.init_std
            cutoff_factor = self.config.init_cutoff_factor
        elif self.config.init_fn == InitFnType.mitchell:
            std = 1 / math.sqrt(self.config.d_model)
            cutoff_factor = self.config.init_cutoff_factor or 3.0
        elif self.config.init_fn == InitFnType.full_megatron:
            std = self.config.init_std
            cutoff_factor = self.config.init_cutoff_factor or 3.0
        else:
            raise NotImplementedError(self.config.init_fn)

        if not self.qargs.use_quantize_model:
            init_normal(self.att_proj, std, cutoff_factor)
            init_normal(self.ff_proj, std, cutoff_factor)
        else:
            init_normal(self.BeforeAttention.att_proj, std, cutoff_factor)
            init_normal(self.MLPResidual.ff_proj, std, cutoff_factor)

    def forward(
        self,
        x: torch.Tensor,
        qx: torch.Tensor,
        sx: torch.Tensor,
        attention_bias: torch.Tensor | None = None,
        layer_past: tuple[torch.Tensor, torch.Tensor] | None = None,
        use_cache: bool = False,
        max_doc_len: int | None = None,
        cu_doc_lens: torch.Tensor | None = None,
    ) -> tuple[torch.Tensor, tuple[torch.Tensor, torch.Tensor] | None]:
        # Get query, key, value projections.
        # shape:
        #  - for regular attn q, k, v: (batch_size, seq_len, d_model)
        #  - for multi-query attn q: (batch_size, seq_len, d_model)
        #                      k, v: (batch_size, seq_len, d_model // n_heads)
        #  - for group query attn q: (batch_size, seq_len, d_model)
        #                      k, v: (batch_size, seq_len, d_model // n_kv_heads)

        # import IPython
        # IPython.embed()

        if self.qargs.use_quantize_model:
            # if False:
            x, qkv = self.BeforeAttention(x, qx, sx)
        else:
            # apply norm before
            h = self.attn_norm(x)

            qkv = self.BeforeAttention.att_proj(h)

        if self.config.clip_qkv is not None:
            qkv.clamp_(min=-self.config.clip_qkv, max=self.config.clip_qkv)

        q, k, v = qkv.split(self.fused_dims, dim=-1)

        # Get attention scores.
        att, cache = self.attention(
            q,
            k,
            v,
            attention_bias,
            layer_past=layer_past,
            use_cache=use_cache,
            max_doc_len=max_doc_len,
            cu_doc_lens=cu_doc_lens,
        )

        # import IPython
        # IPython.embed()
        if self.qargs.use_quantize_model:
            # if False:
            x, qx, sx = self.AfterAttention(x, att)
        else:
            att = self.AfterAttention.attn_out(att)

            # Add attention scores.
            # shape: (B, T, C)
            x = x + self.dropout(att)

        if self.qargs.use_quantize_model:
            # if False:
            x, qx, sx = self.MLPResidual(x, qx, sx)
        else:
            # Add feed-forward projection.
            # shape: (batch_size, seq_len, d_model)
            og_x = x

            x = self.ff_norm(x)

            x = self.MLPResidual.ff_proj(x)

            if self._activation_checkpoint_fn is not None:
                x = self._activation_checkpoint_fn(self.act, x)  # type: ignore
            else:
                x = self.act(x)
            x = self.MLPResidual.ff_out(x)

            x = self.dropout(x)
            x = og_x + x

        # import IPython
        # IPython.embed()

        return x, qx, sx, cache


class CoatOLMoBlockGroup(nn.ModuleList):
    def __init__(self, config: ModelConfig, layer_offset: int, modules: Iterable[nn.Module] | None = None):
        super().__init__(modules)
        self.config = config
        self.layer_offset = layer_offset
        self.activation_checkpointing_strategy: ActivationCheckpointingStrategy | None = None
        self._activation_checkpoint_fn = activation_checkpoint_function(self.config)

    def forward(
        self,
        x: torch.Tensor,
        attention_bias: torch.FloatTensor | None = None,
        layers_past: list[tuple[torch.Tensor, torch.Tensor]] | None = None,
        use_cache: bool = False,
        max_doc_len: int | None = None,
        cu_doc_lens: torch.Tensor | None = None,
    ) -> tuple[torch.Tensor, list[tuple[torch.Tensor, torch.Tensor]] | None]:
        attn_key_values: list[tuple[torch.Tensor, torch.Tensor]] | None = [] if use_cache else None
        for block_idx, block in enumerate(self):
            layer_past = None if layers_past is None else layers_past[block_idx]
            block_idx += self.layer_offset
            if should_checkpoint_block(self.activation_checkpointing_strategy, block_idx):
                # shape: (batch_size, seq_len, d_model)
                x, cache = self._activation_checkpoint_fn(  # type: ignore
                    block,
                    x,
                    attention_bias=attention_bias,
                    layer_past=layer_past,
                    use_cache=use_cache,
                    max_doc_len=max_doc_len,
                    cu_doc_lens=cu_doc_lens,
                )
            else:
                # shape: (batch_size, seq_len, d_model)
                x, cache = block(
                    x,
                    attention_bias=attention_bias,
                    layer_past=layer_past,
                    use_cache=use_cache,
                    max_doc_len=max_doc_len,
                    cu_doc_lens=cu_doc_lens,
                )
            if attn_key_values is not None:
                assert cache is not None
                attn_key_values.append(cache)
        return x, attn_key_values

    def reset_parameters(self):
        for block in self:
            block.reset_parameters()

    def set_activation_checkpointing(
        self, strategy: ActivationCheckpointingStrategy | None, checkpoint_func: Callable | None = None
    ):
        self.activation_checkpointing_strategy = strategy
        for block in self:
            block.set_activation_checkpointing(strategy, checkpoint_func=checkpoint_func)


class CoatOLMo(nn.Module):
    def __init__(self, config: ModelConfig, qargs: QuantActivationConfig, init_params: bool = True):
        super().__init__()
        self.config = config
        self.qargs = qargs
        self.__cache = BufferCache()

        # Validate config.
        if self.config.alibi and self.config.flash_attention:
            raise OLMoConfigurationError("ALiBi is currently not supported with FlashAttention")

        if self.config.alibi and self.config.rope:
            raise OLMoConfigurationError("ALiBi and RoPE are mutually exclusive")

        if self.config.embedding_size is not None and self.config.embedding_size != self.config.vocab_size:
            if self.config.embedding_size < self.config.vocab_size:
                raise OLMoConfigurationError("embedding size should be at least as big as vocab size")
            elif self.config.embedding_size % 128 != 0:
                import warnings

                warnings.warn(
                    "Embedding size is not a multiple of 128! This could hurt throughput performance.", UserWarning
                )

        self.activation_checkpointing_strategy: ActivationCheckpointingStrategy | None = None
        self._activation_checkpoint_fn: Callable = activation_checkpoint_function(self.config)

        if not (
            0 < self.config.block_group_size <= self.config.n_layers
            and self.config.n_layers % self.config.block_group_size == 0
        ):
            raise OLMoConfigurationError("n layers must be divisible by block group size")

        torch.backends.cuda.enable_flash_sdp(True)
        torch.backends.cuda.enable_mem_efficient_sdp(False)  # this is super slow so make sure torch won't use it

        self.transformer = nn.ModuleDict(
            dict(
                wte=nn.Embedding(config.embedding_size or config.vocab_size, config.d_model, device=config.init_device),
                emb_drop=Dropout(config.embedding_dropout),
                ln_f=LayerNorm.build(config),
            )
        )

        blocks = [CoatOLMoBlock.build(i, config, qargs, self.__cache) for i in range(config.n_layers)]
        if self.config.block_group_size > 1:
            block_groups = [
                CoatOLMoBlockGroup(config, i, blocks[i : i + config.block_group_size])
                for i in range(0, config.n_layers, config.block_group_size)
            ]
            self.transformer.update({"block_groups": nn.ModuleList(block_groups)})
        else:
            self.transformer.update({"blocks": nn.ModuleList(blocks)})

        if not (self.config.alibi or self.config.rope):
            self.transformer.update(
                {"wpe": nn.Embedding(config.max_sequence_length, config.d_model, device=config.init_device)}
            )
        if not config.weight_tying:
            self.transformer.update(
                {
                    "ff_out": nn.Linear(
                        config.d_model,
                        config.embedding_size or config.vocab_size,
                        bias=config.include_bias,
                        device=config.init_device,
                    )
                }
            )
        if config.embedding_layer_norm:
            self.transformer.update({"emb_norm": LayerNorm.build(config)})

        # When `init_device="meta"` FSDP will call `reset_parameters()` to initialize weights.
        if init_params and self.config.init_device != "meta":
            self.reset_parameters()
        self.__num_fwd_flops: int | None = None
        self.__num_bck_flops: int | None = None

        # Warm up cache.
        if self.config.alibi:
            get_causal_attention_bias(self.__cache, config.max_sequence_length, _non_meta_init_device(config))
            self.get_alibi_attention_bias(config.max_sequence_length, _non_meta_init_device(config))

        # Quantize
        self.quantize_input_before_block = Coat_quantize_bgn(qargs)
        self.quantize_output_after_block = Coat_quantize_end(qargs)

    set_activation_checkpointing = OLMo.set_activation_checkpointing
    device = OLMo.device
    reset_parameters = OLMo.reset_parameters
    get_alibi_attention_bias = OLMo.get_alibi_attention_bias

    def forward(
        self,
        input_ids: torch.LongTensor,
        input_embeddings: torch.FloatTensor | None = None,
        attention_mask: torch.Tensor | None = None,
        attention_bias: torch.Tensor | None = None,
        past_key_values: Sequence[tuple[torch.Tensor, torch.Tensor]] | None = None,
        use_cache: bool = False,
        last_logits_only: bool = False,
        output_hidden_states: bool | None = None,
        doc_lens: torch.Tensor | None = None,
        max_doc_lens: Sequence[int] | None = None,
    ) -> OLMoOutput:
        """
        :param input_ids: A tensor of shape `(batch_size, seq_len)`.
        :param input_embeddings: A tensor of shape `(batch_size, seq_len, d_model)` with input
            embeddings. When provided, it is treated as the output of the input embedding layer.
        :param attention_mask: A tensor of shape `(batch_size, seq_len)` that indicates
            which input IDs are masked. A `1` value in the mask means that
            the corresponding input ID should *not* be ignored. A `0` means
            that the corresponding input ID is masked.

            This has the same meaning as the `attention_mask` in HuggingFace's `transformers`
            library.
        :param attention_bias: A tensor of shape `(batch_size, 1, seq_len, seq_len)`,
            `(1, 1, seq_len, seq_len)`, or `(seq_len, seq_len)`. This is used
            to introduce causal or other biases.

            If the tensor is a bool or byte tensor, a `True` or `1` at `attention_bias[:, :, i, j]`
            indicates that the i-th element in the sequence is allowed to attend to the j-th
            element in the sequence.

            If the tensor is a float tensor, it will just be added to the attention
            scores before the softmax.

            The default is causal, which corresponds to a lower-diagonal byte matrix of ones.
        :param past_key_values: Pre-computed keys and values for each attention block.
            Can be used to speed up sequential decoding. The `input_ids` which have
            their past given to this model should not be passed as `input_ids` as they have already been computed.
        :param use_cache: If `True`, return key and value tensors for each block.
        :param last_logits_only: If `True`, only compute the logits for the last token of each sequence.
            This can speed up decoding when you only care about the next token.
        :param doc_lens: Document lengths to use in attention for intra-document masking.
            Shape `(batch_size, max_docs)`.
        :param max_doc_lens: Maximum document length for each instance in the batch.
        """
        output_hidden_states = output_hidden_states if output_hidden_states is not None else False

        if past_key_values:
            assert len(past_key_values) == self.config.n_layers

        batch_size, seq_len = input_ids.size() if input_embeddings is None else input_embeddings.size()[:2]
        if past_key_values is None:
            past_length = 0
        else:
            past_length = past_key_values[0][0].size(-2)

        max_doc_len: int | None = None
        cu_doc_lens: torch.Tensor | None = None
        if doc_lens is not None and max_doc_lens is not None:
            max_doc_len = max(max_doc_lens)
            cu_doc_lens = get_cumulative_document_lengths(doc_lens)

        # Get embeddings of input.
        # shape: (batch_size, seq_len, d_model)
        x = self.transformer.wte(input_ids) if input_embeddings is None else input_embeddings  # type: ignore

        # Apply embedding layer norm.
        if self.config.embedding_layer_norm:
            x = self.transformer.emb_norm(x)

        if not (self.config.alibi or self.config.rope):
            # Get positional embeddings.
            # shape: (1, seq_len)
            pos = torch.arange(past_length, past_length + seq_len, dtype=torch.long, device=x.device).unsqueeze(0)
            # shape: (1, seq_len, d_model)
            pos_emb = self.transformer.wpe(pos)  # type: ignore
            x = pos_emb + x

        # Apply dropout.
        # shape: (batch_size, seq_len, d_model)
        x = self.transformer.emb_drop(x)  # type: ignore

        # Transform the attention mask into what the blocks expect.
        if attention_mask is not None:
            # shape: (batch_size, 1, 1, seq_len)
            attention_mask = attention_mask.to(dtype=torch.float).view(batch_size, -1)[:, None, None, :]
            attention_mask = (1.0 - attention_mask) * torch.finfo(attention_mask.dtype).min

        # Merge attention mask with attention bias.
        if (
            attention_bias is not None
            or attention_mask is not None
            or self.config.alibi
            # NOTE (epwalsh): we need to initialize the attn bias in order for attn to work properly
            # with key+value cache. Otherwise `F.scaled_dot_product_attention()` doesn't seem to compute
            # scores correctly.
            or past_key_values is not None
        ):
            if attention_bias is None and self.config.alibi:
                attention_bias = get_causal_attention_bias(
                    self.__cache, past_length + seq_len, x.device
                ) + self.get_alibi_attention_bias(past_length + seq_len, x.device)
            elif attention_bias is None:
                attention_bias = get_causal_attention_bias(self.__cache, past_length + seq_len, x.device)
            elif attention_bias.dtype in (torch.int8, torch.bool):
                attention_bias = attention_bias.to(dtype=torch.float)
                attention_bias.masked_fill_(attention_bias == 0.0, torch.finfo(attention_bias.dtype).min)

            # Transform to the right shape and data type.
            mask_len = seq_len
            if attention_mask is not None:
                mask_len = attention_mask.shape[-1]
            elif past_key_values is not None:
                mask_len = past_key_values[0][0].shape[-2] + seq_len
            attention_bias = attention_bias[:, :, :mask_len, :mask_len].to(dtype=torch.float)

            # Add in the masking bias.
            if attention_mask is not None:
                attention_bias = attention_bias + attention_mask
                # Might get -infs after adding attention mask, since dtype.min + dtype.min = -inf.
                # `F.scaled_dot_product_attention()` doesn't handle -inf like you'd expect, instead
                # it can produce NaNs.
                ensure_finite_(attention_bias, check_neg_inf=True, check_pos_inf=False)

        attn_key_values: list[tuple[torch.Tensor, torch.Tensor]] | None = [] if use_cache else None

        # decoder layers
        all_hidden_states = []

        # Prepare the input for COAT decoderlayer
        x, qx, sx = self.quantize_input_before_block(x)

        # Apply blocks one-by-one.
        if self.config.block_group_size == 1:
            for block_idx, block in enumerate(self.transformer.blocks):
                if output_hidden_states:
                    # add hidden states
                    all_hidden_states.append(x)

                layer_past = None if past_key_values is None else past_key_values[block_idx]
                if should_checkpoint_block(self.activation_checkpointing_strategy, block_idx):
                    # shape: (batch_size, seq_len, d_model)
                    x, qx, sx, cache = self._activation_checkpoint_fn(
                        block,
                        x,
                        qx,
                        sx,
                        attention_bias=attention_bias,
                        layer_past=layer_past,
                        use_cache=use_cache,
                        max_doc_len=max_doc_len,
                        cu_doc_lens=cu_doc_lens,
                    )
                else:
                    # shape: (batch_size, seq_len, d_model)
                    x, qx, sx, cache = block(
                        x,
                        qx,
                        sx,
                        attention_bias=attention_bias,
                        layer_past=layer_past,
                        use_cache=use_cache,
                        max_doc_len=max_doc_len,
                        cu_doc_lens=cu_doc_lens,
                    )

                if attn_key_values is not None:
                    assert cache is not None
                    attn_key_values.append(cache)
        else:
            for group_idx, block_group in enumerate(self.transformer.block_groups):
                if output_hidden_states:
                    # add hidden states
                    all_hidden_states.append(x)

                layers_past = (
                    None
                    if past_key_values is None
                    else past_key_values[
                        group_idx * self.config.block_group_size : (group_idx + 1) * self.config.block_group_size
                    ]
                )
                x, cache = block_group(
                    x,
                    attention_bias=attention_bias,
                    layers_past=layers_past,
                    use_cache=use_cache,
                    max_doc_len=max_doc_len,
                    cu_doc_lens=cu_doc_lens,
                )
                if attn_key_values is not None:
                    assert cache is not None
                    attn_key_values.extend(cache)

        # Summarize the output of the Decoder Layer
        x = self.quantize_output_after_block(x, qx, sx)

        if last_logits_only:
            # shape: (batch_size, 1, d_model)
            x = x[:, -1, :].unsqueeze(1)

        # Apply final layer norm.
        # shape: (batch_size, seq_len or 1, d_model)
        x = self.transformer.ln_f(x)  # type: ignore
        if output_hidden_states:
            # add final hidden state post-final-layernorm, following HuggingFace's convention
            all_hidden_states.append(x)

        # Get logits.
        # shape: (batch_size, seq_len or 1, vocab_size)
        if self.config.weight_tying:
            logits = F.linear(x, self.transformer.wte.weight, None)  # type: ignore
        else:
            logits = self.transformer.ff_out(x)  # type: ignore
        if self.config.scale_logits:
            logits.mul_(1 / math.sqrt(self.config.d_model))

        return OLMoOutput(
            logits=logits,
            attn_key_values=attn_key_values,
            hidden_states=tuple(all_hidden_states) if output_hidden_states else None,
        )

    def get_fsdp_wrap_policy(self, wrap_strategy: FSDPWrapStrategy | None = None):
        if wrap_strategy is None:
            return None

        # The 'recurse' mode for the wrap function does not behave like you'd expect.
        # Even if we return False, it may still recurse because PyTorch does what it wants,
        # not what you want. This causes issues when, for example, we want to wrap 'ff_out' (a linear layer)
        # but not other linear layers within a block.
        # So we have to explicitly tell PyTorch which linear layers to wrap, and we also just
        # return True in 'recurse' mode for simplicity.
        size_based_module_to_wrap = {self.transformer.wte}
        if hasattr(self.transformer, "ff_out"):
            size_based_module_to_wrap.add(self.transformer.ff_out)

        if wrap_strategy == FSDPWrapStrategy.by_block:

            def fsdp_wrap_fn(module, recurse: bool = True, nonwrapped_numel: int = 0):
                del nonwrapped_numel
                wrap = isinstance(module, CoatOLMoBlock)
                if recurse:
                    return True
                else:
                    return wrap

            return fsdp_wrap_fn
        elif wrap_strategy == FSDPWrapStrategy.by_block_and_size:

            def fsdp_wrap_fn(module, recurse: bool = True, nonwrapped_numel: int = 0):
                del nonwrapped_numel
                wrap = isinstance(module, (CoatOLMoBlock,)) or module in size_based_module_to_wrap
                if recurse:
                    return True
                else:
                    return wrap

            return fsdp_wrap_fn
        elif wrap_strategy == FSDPWrapStrategy.by_block_group:
            if self.config.block_group_size <= 1:
                raise OLMoConfigurationError(
                    "'by_block_group' FSDP wrapping strategy requires block group size greater than 1"
                )

            def fsdp_wrap_fn(module, recurse: bool = True, nonwrapped_numel: int = 0):
                del nonwrapped_numel
                wrap = isinstance(module, CoatOLMoBlockGroup)
                if recurse:
                    return True
                else:
                    return wrap

            return fsdp_wrap_fn
        elif wrap_strategy == FSDPWrapStrategy.by_block_group_and_size:
            if self.config.block_group_size <= 1:
                raise OLMoConfigurationError(
                    "'by_block_group_and_size' FSDP wrapping strategy requires block group size greater than 1"
                )

            def fsdp_wrap_fn(module, recurse: bool = True, nonwrapped_numel: int = 0):
                del nonwrapped_numel
                wrap = isinstance(module, (CoatOLMoBlockGroup,)) or module in size_based_module_to_wrap
                if recurse:
                    return True
                else:
                    return wrap

            return fsdp_wrap_fn
        elif wrap_strategy == FSDPWrapStrategy.size_based:
            from torch.distributed.fsdp.wrap import size_based_auto_wrap_policy

            return size_based_auto_wrap_policy
        elif wrap_strategy in {
            FSDPWrapStrategy.one_in_two,
            FSDPWrapStrategy.one_in_three,
            FSDPWrapStrategy.one_in_four,
            FSDPWrapStrategy.one_in_five,
        }:
            c = {
                FSDPWrapStrategy.one_in_two: 2,
                FSDPWrapStrategy.one_in_three: 3,
                FSDPWrapStrategy.one_in_four: 4,
                FSDPWrapStrategy.one_in_five: 5,
            }[wrap_strategy]

            def fsdp_wrap_fn(module, recurse: bool = True, nonwrapped_numel: int = 0):
                del nonwrapped_numel
                wrap = isinstance(module, CoatOLMoBlock) and module.layer_id % c == 0
                if recurse:
                    return True
                else:
                    return wrap

            return fsdp_wrap_fn
        else:
            raise NotImplementedError(wrap_strategy)

    num_params = OLMo.num_params

    @property
    def num_fwd_flops(self):
        if self.__num_fwd_flops:
            return self.__num_fwd_flops

        # embedding table is just a lookup in the forward pass
        n_params = self.num_params(include_embedding=False)
        # the number of parameters is approximately the number of multiply-accumulates (MAC) in the network
        # each MAC has 2 FLOPs - we multiply by 2 ie 2 * n_param
        # this gets us FLOPs / token
        params_flops_per_token = 2 * n_params
        # there are 2 FLOPS per mac; there is A=Q*K^T and out=A*V ops (ie mult by 2)
        attn_flops_per_token = self.config.n_layers * 2 * 2 * (self.config.d_model * self.config.max_sequence_length)
        self.__num_fwd_flops = params_flops_per_token + attn_flops_per_token
        return self.__num_fwd_flops

    @property
    def num_bck_flops(self):
        if self.__num_bck_flops:
            return self.__num_bck_flops

        n_params = self.num_params()
        params_flops_per_token = 4 * n_params
        attn_flops_per_token = self.config.n_layers * 8 * (self.config.d_model * self.config.max_sequence_length)
        self.__num_bck_flops = params_flops_per_token + attn_flops_per_token
        return self.__num_bck_flops

    generate = OLMo.generate

    @classmethod
    def from_checkpoint(
        cls, checkpoint_dir: PathOrStr, device: str = "cpu", checkpoint_type: CheckpointType | None = None
    ) -> CoatOLMo:
        """
        Load an OLMo model from a checkpoint.
        """
        from olmo.util import resource_path

        # Guess checkpoint type.
        if checkpoint_type is None:
            try:
                if resource_path(checkpoint_dir, "model.pt").is_file():
                    checkpoint_type = CheckpointType.unsharded
                else:
                    checkpoint_type = CheckpointType.sharded
            except FileNotFoundError:
                checkpoint_type = CheckpointType.sharded

        # Load config.
        config_path = resource_path(checkpoint_dir, "config.yaml")
        model_config = ModelConfig.load(config_path, key="model", validate_paths=False)

        if checkpoint_type == CheckpointType.unsharded:
            # Initialize model (always on CPU to start with so we don't run out of GPU memory).
            model_config.init_device = "cpu"
            model = CoatOLMo(model_config)

            # Load state dict directly to target device.
            state_dict_path = resource_path(checkpoint_dir, "model.pt")
            state_dict = torch.load(state_dict_path, map_location="cpu")
            model.load_state_dict(model._make_state_dict_compatible(state_dict)[0])
            model = model.to(torch.device(device))
        else:
            train_config = TrainConfig.load(config_path)
            if train_config.sharded_checkpointer == ShardedCheckpointerType.olmo_core:
                from olmo_core.distributed.checkpoint import \
                    load_model_and_optim_state  # type: ignore

                model_config.init_device = device
                model = CoatOLMo(model_config)
                load_model_and_optim_state(checkpoint_dir, model)
            else:
                # train_config.sharded_checkpointer == ShardedCheckpointerType.torch_new
                from olmo.checkpoint import load_model_state

                # Initialize model on target device. In this case the state dict is loaded in-place
                # so it's not necessary to start on CPU if the target device is a GPU.
                model_config.init_device = device
                model = CoatOLMo(model_config)

                # Load state dict in place.
                load_model_state(checkpoint_dir, model)

        return model.eval()

    def _make_state_dict_compatible(
        self, state_dict: dict[str, torch.Tensor]
    ) -> tuple[dict[str, torch.Tensor], dict[str, set[str]]]:
        """
        Handles some cases where the state dict is valid yet may need to be transformed in order to
        be loaded.

        This modifies the state dict in-place and also returns it, along with a mapping of original key
        names to new key names in cases where the keys were simply renamed. That mapping can be used
        to make a corresponding optimizer state dict compatible as well.
        """
        import re
        from fnmatch import fnmatch

        new_keys_to_og_keys: dict[str, str] = {}

        # Remove "_fsdp_wrapped_module." prefix from all keys. We don't want this prefix when the model is
        # not wrapped in FSDP. And when the model is wrapped in FSDP, loading this state dict will still work
        # fine without the prefixes. This also simplifies the other steps below.
        for key in list(state_dict.keys()):
            state_dict[(new_key := key.replace("_fsdp_wrapped_module.", ""))] = state_dict.pop(key)
            new_keys_to_og_keys[new_key] = key

        # For backwards compatibility prior to fixing https://github.com/allenai/LLM/issues/222
        if self.config.block_type == BlockType.sequential:
            for key in list(state_dict.keys()):
                if fnmatch(key, "transformer.*.norm.weight"):
                    tensor = state_dict.pop(key)
                    state_dict[(new_key := key.replace("norm.weight", "attn_norm.weight"))] = tensor
                    new_keys_to_og_keys[new_key] = new_keys_to_og_keys[key]
                    state_dict[(new_key := key.replace("norm.weight", "ff_norm.weight"))] = tensor.clone()
                    new_keys_to_og_keys[new_key] = new_keys_to_og_keys[key]
                    del new_keys_to_og_keys[key]
                elif fnmatch(key, "transformer.*.norm.bias"):
                    tensor = state_dict.pop(key)
                    state_dict[(new_key := key.replace("norm.bias", "attn_norm.bias"))] = tensor
                    new_keys_to_og_keys[new_key] = new_keys_to_og_keys[key]
                    state_dict[(new_key := key.replace("norm.bias", "ff_norm.bias"))] = tensor.clone()
                    new_keys_to_og_keys[new_key] = new_keys_to_og_keys[key]
                    del new_keys_to_og_keys[key]

        # Realquantization will change the place the linear layers happen
        if self.qargs.use_quantize_model == "coat_real":
            for key in list(state_dict.keys()):
                if fnmatch(key, "transformer.blocks.*.att_proj.weight") and "BeforeAttention" not in key:
                    tensor = state_dict.pop(key)
                    state_dict[(new_key := key.replace("att_proj.weight", "BeforeAttention.att_proj.weight"))] = tensor
                    new_keys_to_og_keys[new_key] = new_keys_to_og_keys[key]
                    del new_keys_to_og_keys[key]
                elif fnmatch(key, "transformer.blocks.*.attn_out.weight") and "AfterAttention" not in key:
                    tensor = state_dict.pop(key)
                    state_dict[(new_key := key.replace("attn_out.weight", "AfterAttention.attn_out.weight"))] = tensor
                    new_keys_to_og_keys[new_key] = new_keys_to_og_keys[key]
                    del new_keys_to_og_keys[key]
                elif fnmatch(key, "transformer.blocks.*.ff_proj.weight") and "MLPResidual" not in key:
                    tensor = state_dict.pop(key)
                    state_dict[(new_key := key.replace("ff_proj.weight", "MLPResidual.ff_proj.weight"))] = tensor
                    new_keys_to_og_keys[new_key] = new_keys_to_og_keys[key]
                    del new_keys_to_og_keys[key]
                elif fnmatch(key, "transformer.blocks.*.ff_out.weight") and "MLPResidual" not in key:
                    tensor = state_dict.pop(key)
                    state_dict[(new_key := key.replace("ff_out.weight", "MLPResidual.ff_out.weight"))] = tensor
                    new_keys_to_og_keys[new_key] = new_keys_to_og_keys[key]
                    del new_keys_to_og_keys[key]

        # For loading a state dict that was saved with a different `block_group_size`.
        if "transformer.block_groups.0.0.attn_out.weight" in state_dict.keys():
            state_dict_block_group_size = len(
                [k for k in state_dict.keys() if fnmatch(k, "transformer.block_groups.0.*.attn_out.weight")]
            )
        else:
            state_dict_block_group_size = 1
        if self.config.block_group_size != state_dict_block_group_size:
            log.info(
                f"Regrouping state dict blocks from group size {state_dict_block_group_size} to "
                f"group size {self.config.block_group_size}"
            )
            # For simplicity we're first going to flatten out the block groups in the state dict (if necessary)
            # and then (re-)group them into the right block sizes.
            if state_dict_block_group_size > 1:
                for key in list(state_dict.keys()):
                    if (m := re.match(r"transformer.block_groups\.(\d+)\.(\d+)\..*", key)) is not None:
                        group_idx, group_block_idx = int(m.group(1)), int(m.group(2))
                        block_idx = (group_idx * state_dict_block_group_size) + group_block_idx
                        state_dict[
                            (
                                new_key := key.replace(
                                    f"block_groups.{group_idx}.{group_block_idx}.", f"blocks.{block_idx}."
                                )
                            )
                        ] = state_dict.pop(key)
                        new_keys_to_og_keys[new_key] = new_keys_to_og_keys.pop(key)

            if self.config.block_group_size > 1:
                # Group the state dict blocks into the right block size.
                for key in list(state_dict.keys()):
                    if (m := re.match(r"transformer.blocks\.(\d+)\..*", key)) is not None:
                        block_idx = int(m.group(1))
                        group_idx, group_block_idx = (
                            block_idx // self.config.block_group_size,
                            block_idx % self.config.block_group_size,
                        )
                        state_dict[
                            (
                                new_key := key.replace(
                                    f"blocks.{block_idx}.", f"block_groups.{group_idx}.{group_block_idx}."
                                )
                            )
                        ] = state_dict.pop(key)
                        new_keys_to_og_keys[new_key] = new_keys_to_og_keys.pop(key)

        og_keys_to_new: dict[str, set[str]] = defaultdict(set)
        for new_key, og_key in new_keys_to_og_keys.items():
            og_keys_to_new[og_key].add(new_key)

        return state_dict, og_keys_to_new



import os
import torch
import argparse
import transformers
from transformers import AutoModelForCausalLM, AutoConfig
from typing import Optional
from dataclasses import dataclass, field, asdict

from .coat_llama import make_state_dict_compatible, CoatLlamaForCausalLM, CoatLlamaConfig
from ..utils._fp8_quantization_config import QuantizationConfig


@dataclass
class ConvertArguments:
    model_name: str = field(metadata={"help": "The model name or path to download the LLaMA model"})
    save_path: str = field(metadata={"help": "The path where the converted model weights will be saved"})
    cache_dir: str = field(default=None, metadata={"help": "Directory to cache the model"})


def download_and_convert_llama(convert_args: ConvertArguments, quantization_args: QuantizationConfig):
    """
    Downloads a LLaMA model, converts its weights using `make_state_dict_compatible`,
    and saves the converted model.

    Args:
        model_name (str): The model name or path to download the LLaMA model.
        save_path (str): The path where the converted model weights will be saved.
        cache_dir (Optional[str]): Directory to cache the model. Defaults to None.

    Returns:
        None
    """
    model_name = convert_args.model_name
    save_path = convert_args.save_path
    cache_dir = convert_args.cache_dir
    
    # Step 1: Download the original LLaMA model
    model = AutoModelForCausalLM.from_pretrained(
        model_name,
        cache_dir=cache_dir
    )
    
    # Step 2: Initialize the model configuration for FP8 or other custom config
    config = AutoConfig.from_pretrained(
        model_name,
        cache_dir=cache_dir
    )
    
    # Step 3: Apply make_state_dict_compatible to convert weights
    compatible_state_dict = make_state_dict_compatible(model.state_dict())
    
    # Step 4: Create a new model instance with compatible configuration
    fp8_config = CoatLlamaConfig(**config.to_dict())
    fp8_config.coat_fp8_args = asdict(quantization_args)

    converted_model = AutoModelForCausalLM.from_config(fp8_config)
    converted_model.load_state_dict(compatible_state_dict)
    
    # Step 5: Save the converted model and configuration using save_pretrained
    os.makedirs(save_path, exist_ok=True)
    converted_model.save_pretrained(save_path)
    print(f"Converted model saved at {save_path}")


if __name__ == "__main__":
    # Parse command-line arguments
    parser = transformers.HfArgumentParser( # NOTE: FP8
        (ConvertArguments, QuantizationConfig)
    )
    convert_args, quantization_args = parser.parse_args_into_dataclasses()
    
    print(quantization_args)
    
    # Call the function with parsed arguments
    download_and_convert_llama(
        convert_args, quantization_args
    )


# Copyright 2024 NVIDIA CORPORATION & AFFILIATES
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0

# This repo is used to reproduce Table 6 in COAT paper

import pandas as pd
import torch
import argparse
from transformers import AutoConfig, AutoModel
from transformers.models.llama.modeling_llama import LlamaDecoderLayer

from coat.activation.models._fp8manager import FP8Manager
from coat.activation.models._fp8_quantization_config import QuantizationConfig
from coat.activation.models.coat_llama import CoatLlamaDecoderLayer, CoatLlamaModel, make_state_dict_compatible
from coat.activation.real_quantization import (Coat_quantize_bgn,
                                               Coat_quantize_end)

torch.manual_seed(0)
batch_sizes = [8]
sequence_lengths = [2048]
n_repeat = 20

torch.empty(
    1, device="cuda", requires_grad=True
).backward()  # Triton will throw RuntimeError: Triton Error [CUDA]: invalid device context if you comment this line

config = AutoConfig.from_pretrained("meta-llama/Llama-3.1-8B")
config._attn_implementation = "flash_attention_2"
model = AutoModel.from_pretrained("meta-llama/Llama-3.1-8B")

def tester(CHOICE):
    if CHOICE == "BF16":
        llama_layer = LlamaDecoderLayer(config=config, layer_idx=0)
        llama_layer = llama_layer.cuda().to(torch.bfloat16)
        llama_layer.load_state_dict(model.layers[0].state_dict())

        # Forward Time Benchmark
        for batch_size in batch_sizes:
            for sequence_length in sequence_lengths:
                dummy_input = (
                    torch.rand((batch_size, sequence_length, config.hidden_size), dtype=torch.bfloat16)
                    .cuda()
                    .requires_grad_(True)
                )

                attention_mask = torch.ones((batch_size, sequence_length), dtype=torch.bool).cuda()
                position_ids = torch.arange(sequence_length, dtype=torch.long, device="cuda").unsqueeze(0).expand(batch_size, -1)

                dummy_grad = torch.rand((batch_size, sequence_length, config.hidden_size), dtype=torch.bfloat16).cuda()

                # Test Backward Time
                with torch.autocast(device_type="cuda", dtype=torch.bfloat16):
                    dummy_output = llama_layer(dummy_input, attention_mask, position_ids)
                    dummy_output[0].backward(dummy_grad)
                    torch.cuda.synchronize()
                    import IPython
                    IPython.embed()

    if CHOICE == "COAT":
        fp8_config = QuantizationConfig(
            quantize_model=True,
            group_size=16,
        )
        config.coat_fp8_args = fp8_config

        llama_layer = CoatLlamaDecoderLayer(config, layer_idx=0)
        llama_layer.load_state_dict(make_state_dict_compatible(model.layers[0].state_dict()))

        llama_bgn = Coat_quantize_bgn(config.coat_fp8_args).cuda()
        llama_end = Coat_quantize_end(config.coat_fp8_args).cuda()
        llama_layer = llama_layer.cuda().to(torch.bfloat16)

        # Forward Time Benchmark
        for batch_size in batch_sizes:
            for sequence_length in sequence_lengths:
                dummy_input = (
                    torch.rand((batch_size, sequence_length, config.hidden_size), dtype=torch.bfloat16)
                    .cuda()
                    .requires_grad_(True)
                )
                
                attention_mask = torch.ones((batch_size, sequence_length), dtype=torch.bool).cuda()
                position_ids = torch.arange(sequence_length, dtype=torch.long, device="cuda").unsqueeze(0).expand(batch_size, -1)

                dummy_grad = torch.rand((batch_size, sequence_length, config.hidden_size), dtype=torch.bfloat16).cuda()

                dummy_x, dummy_qx, dummy_sx = llama_bgn(dummy_input)

                # Test Backward Time
                with torch.autocast(device_type="cuda", dtype=torch.bfloat16):
                    FP8Manager.is_first_microbatch = True
                    dummy_output_x, dummy_output_qx, dummy_output_sx = llama_layer(
                        dummy_x, dummy_qx, dummy_sx, attention_mask, position_ids
                    )
                    dummy_output = llama_end(dummy_output_x, dummy_output_qx, dummy_output_sx)
                    dummy_output.backward(dummy_grad)
                    torch.cuda.synchronize()
                    import IPython
                    IPython.embed()

def parse_arguments():
    parser = argparse.ArgumentParser(description="Set the method for training.")
    parser.add_argument(
        "--method",
        type=str,
        choices=["COAT", "BF16"],
        required=True,
        help="Choose the training method: 'COAT' or 'BF16'."
    )
    return parser.parse_args()

torch.manual_seed(0)
torch.set_printoptions(linewidth=200, sci_mode=False)
args = parse_arguments()
tester(args.method)


# Copyright 2024 NVIDIA CORPORATION & AFFILIATES
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0

import unittest

import torch
import triton
from coat.activation.real_quantization.linear import fp8matmul
from coat.activation.real_quantization.fp8linear import FP8Linear
from coat.activation.deepseek.fp8linear import FP8DeepSeekLinear
from .test_utils import check_similarity, dequantize_tensor, quantize_tensor


def _test_linear_mse(input, weight, gradient, m, n, k, device):
    with torch.autocast(device_type="cuda", dtype=torch.bfloat16):
        # Forward without Quantization
        bf16_input = input.clone().requires_grad_(True)
        bf16_linear = torch.nn.Linear(k, n, bias=False).to(torch.bfloat16)
        bf16_linear.weight.data = weight
        output_bf16_fwd = bf16_linear(bf16_input)

        # Forward + Quantization
        bf16_input_quant = input.clone().requires_grad_(True)
        bf16_linear_quant = FP8Linear(k, n, bias=False, device=device).to(torch.bfloat16)
        bf16_linear_quant.weight.data = weight
        output_bf16_quant_fwd = bf16_linear_quant(bf16_input_quant)

        # Forward + DeepSeek FP8
        fp8_input = input.clone().requires_grad_(True)
        fp8_linear = FP8DeepSeekLinear(k, n, bias=False, device=device).to(torch.bfloat16)
        fp8_linear.weight.data = weight
        output_fp8_fwd = fp8_linear(fp8_input)

        # Backward
        output_bf16_fwd.backward(gradient, retain_graph=True)
        output_bf16_quant_fwd.backward(gradient, retain_graph=True)
        output_fp8_fwd.backward(gradient, retain_graph=True)

        grad_input_bf16_bwd = bf16_input.grad.clone()
        grad_weight_bf16_bwd = bf16_linear.weight.grad.clone()
        grad_input_bf16_quant_bwd = bf16_input_quant.grad.clone()
        grad_weight_bf16_quant_bwd = bf16_linear_quant.weight.grad.clone()
        grad_input_fp8_bwd = fp8_input.grad.clone()
        grad_weight_fp8_bwd = fp8_linear.weight.grad.clone()

    return (
        output_bf16_fwd, output_bf16_quant_fwd, output_fp8_fwd, 
        grad_input_bf16_bwd, grad_input_bf16_quant_bwd, grad_input_fp8_bwd, 
        grad_weight_bf16_bwd, grad_weight_bf16_quant_bwd, grad_weight_fp8_bwd
    )
    
def _test_deepseek_linear(input, weight, gradient, m, n, k, device):
    bf16_input = input.clone().requires_grad_(True)
    fp8_input = input.clone().requires_grad_(True)
    
    # # Forward without Quantization
    # bf16_linear = torch.nn.Linear(k, n, bias=False).to(torch.bfloat16)
    # bf16_linear.weight.data = weight
    # output_bf16_fwd = bf16_linear(fp8_input)

    # Forward + Quantization
    bf16_linear = torch.nn.Linear(k, n, bias=False).to(torch.bfloat16)
    quant_input = quantize_tensor(input, 1, m, k, 128, torch.float8_e4m3fn, quant_type="per_group")
    quant_weight = quantize_tensor(weight, 1, n, k, 128, torch.float8_e4m3fn, quant_type="per_block")
    bf16_linear.weight.data = quant_weight[0]
    bf16_input = quant_input[0].squeeze(0).clone().detach().requires_grad_(True)
    output_bf16_fwd = bf16_linear(bf16_input)
    
    fp8_linear = FP8DeepSeekLinear(k, n, bias=False, device=device).to(torch.bfloat16)
    fp8_linear.weight.data = weight
    output_fp8_fwd = fp8_linear(fp8_input)

    # Backward
    output_bf16_fwd.backward(gradient)
    
    output_fp8_fwd.backward(gradient)

    grad_input_bf16_bwd = bf16_input.grad
    grad_weight_bf16_bwd = bf16_linear.weight.grad
    
    grad_input_fp8_bwd = fp8_input.grad
    grad_weight_fp8_bwd = fp8_linear.weight.grad
    
    return (
        output_bf16_fwd, output_fp8_fwd, 
        grad_input_bf16_bwd, grad_input_fp8_bwd, 
        grad_weight_bf16_bwd, grad_weight_fp8_bwd
    )


class TestLinear(unittest.TestCase):
    def test_linear(self):
        M, N, K = 4096, 2048, 8192
        QB = 16
        a = torch.randn((M, K), device="cuda", dtype=torch.bfloat16)
        b = torch.randn((N, K), device="cuda", dtype=torch.bfloat16)

        scale_a, scale_b = torch.randn((1), device="cuda", dtype=torch.bfloat16), torch.randn(
            (1), device="cuda", dtype=torch.bfloat16
        )

        # Prepare row major and col-major data
        a = a.to(torch.float8_e4m3fn)
        b = b.T
        b = b.to(torch.float8_e4m3fn)

        a_32, b_32 = a.to(torch.float32), b.to(torch.float32)
        scale_ab = scale_a.to(torch.float32) * scale_b.to(torch.float32)
        output_torch = torch.matmul(a_32, b_32) * scale_ab
        output_torch_quantized, _, __ = quantize_tensor(output_torch.unsqueeze(0), 1, M, N, QB, torch.float8_e4m3fn)
        output_torch = output_torch.to(torch.bfloat16)

        # Output is not quantized
        output_fp8 = fp8matmul(a, b, False, scale_a, scale_b, QB)  # a should be row-major, b should be col-major

        # Output is quantized
        output_fp8_qx, output_fp8_sx = fp8matmul(
            a, b, True, scale_a, scale_b, QB
        )  # a should be row-major, b should be col-major
        output_fp8_rqx = dequantize_tensor(
            output_fp8_qx.unsqueeze(0), output_fp8_sx.unsqueeze(0), 1, M, N, QB
        )  # dequantize per-tensor quantization

        # TODO: This is not ideal
        # self.assertTrue(torch.allclose(output_torch, output_fp8, 1e-1, 0.2))
        # self.assertTrue(torch.allclose(output_torch_quantized, output_fp8_rqx, 1e-1, 0.5))

    def test_deepseek_linear(self):
        M, N, K = 4096, 2048, 8192
        device, QB = "cuda", 128
        
        a = torch.randn((M, K), device=device, dtype=torch.bfloat16)
        b = torch.randn((N, K), device=device, dtype=torch.bfloat16)
        g = torch.randn((M, N), device=device, dtype=torch.bfloat16)
        
        output_bf16_fwd, output_fp8_deepseek_fwd, grad_input_bf16_bwd, grad_input_fp8_deepseek_bwd, grad_weight_bf16_bwd, grad_weight_fp8_deepseek_bwd = _test_deepseek_linear(a, b, g, M, N, K, device)
        
    def test_linear_mse(self):
        M, N, K = 16384, 16384, 16384
        device, QB = "cuda", 128
        
        a = torch.randn((M, K), device=device, dtype=torch.bfloat16)
        b = torch.randn((N, K), device=device, dtype=torch.bfloat16)
        g = torch.randn((M, N), device=device, dtype=torch.bfloat16)
        
        (
            output_bf16_fwd, output_bf16_quant_fwd, output_fp8_fwd, \
            grad_input_bf16_bwd, grad_input_bf16_quant_bwd, grad_input_fp8_bwd, \
            grad_weight_bf16_bwd, grad_weight_bf16_quant_bwd, grad_weight_fp8_bwd
        ) = _test_linear_mse(a, b, g, M, N, K, device)

        # Calculate MSE
        mse_input_quant = torch.nn.MSELoss()(output_bf16_fwd, output_bf16_quant_fwd)
        mse_input_fp8 = torch.nn.MSELoss()(output_bf16_fwd, output_fp8_fwd)
        mse_grad_input_quant = torch.nn.MSELoss()(grad_input_bf16_bwd, grad_input_bf16_quant_bwd)
        mse_grad_input_fp8 = torch.nn.MSELoss()(grad_input_bf16_bwd, grad_input_fp8_bwd)
        mse_grad_weight_quant = torch.nn.MSELoss()(grad_weight_bf16_bwd, grad_weight_bf16_quant_bwd)
        mse_grad_weight_fp8 = torch.nn.MSELoss()(grad_weight_bf16_bwd, grad_weight_fp8_bwd)
        
        # Print MSEs
        print(f"MSE Input Quant: {mse_input_quant}")
        print(f"MSE Input FP8: {mse_input_fp8}")
        print(f"MSE Grad Input Quant: {mse_grad_input_quant}")
        print(f"MSE Grad Input FP8: {mse_grad_input_fp8}")
        print(f"MSE Grad Weight Quant: {mse_grad_weight_quant}")
        print(f"MSE Grad Weight FP8: {mse_grad_weight_fp8}")

        
if __name__ == "__main__":
    torch.manual_seed(0)
    torch.set_printoptions(sci_mode=False, linewidth=200)
    unittest.main()



# Copyright 2024 NVIDIA CORPORATION & AFFILIATES
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0

"""
Layer Normalization
====================
In this tutorial, you will write a high-performance layer normalization
kernel that runs faster than the PyTorch implementation.

In doing so, you will learn about:

* Implementing backward pass in Triton.

* Implementing parallel reduction in Triton.

"""

import torch
import triton
import triton.language as tl

from ._division import _stochastic_rounding
from ._division_transpose import fp8_division_transpose
from .common import FP8_MAX_VALUE, SCALE_MIN_THRES, convert_fp8_to_embit

"""FP8 LayerNorm. Forward + Backward"""
"""Forward: input uses 1 * 16 quantization"""
"""Forward: output use per-tensor quantization."""
"""Backward: input uses full-precision/BF16."""
"""Backward: output uses full-precision/BF16"""
"""Support 3D input, but need to first flatten to 2D to perform calculation."""


@triton.heuristics(
    {
        "BLOCK_SN": lambda args: args["N"] // args["QB"],
        "BLOCK_SN2": lambda args: args["N2"] // args["QB"],
    }
)
@triton.jit
def _layer_norm_fwd_fused(
    X,  # pointer to the input
    SX,  # pointer to the scale of input
    Y,  # pointer to the output
    SY,  # pointer to the scale of output
    Mean,  # pointer to the mean
    Rstd,  # pointer to the 1/std
    stride,  # how much to increase the pointer when moving by 1 row
    scale_stride,  # how much to increase the pointer when moving by 1 row
    N: tl.constexpr,  # number of columns in X,
    N2: tl.constexpr,  # number of columns in X,
    SN: tl.constexpr,
    SN2: tl.constexpr,
    QB: tl.constexpr,
    eps,  # epsilon to avoid division by zero
    fp8_max,
    SCALE_MIN_THRES: tl.constexpr,
    BLOCK_SN: tl.constexpr,
    BLOCK_SN2: tl.constexpr,
):
    # Map the program id to the row of X and Y it should compute.
    row = tl.program_id(0)
    Y += row * stride
    X += row * stride
    SX += row * scale_stride

    mean = 0
    cols = tl.arange(0, N2)
    scale_cols = tl.arange(0, SN2)
    x = tl.load(X + cols, mask=cols < N, other=0.0).to(tl.float32)
    scale_x = tl.load(SX + scale_cols, mask=scale_cols < SN, other=0.0).to(tl.float32)

    # Dequantize and swish calculation
    scale_x = tl.reshape(scale_x, (BLOCK_SN2, 1))
    x = tl.reshape(x, (BLOCK_SN2, QB))
    x = x * scale_x
    x = tl.reshape(x, (N2,))

    # Compute mean and Variance
    mean = tl.sum(x, axis=0) / N
    # Compute variance
    _var = (x - mean) * (x - mean)
    var = tl.sum(_var, axis=0) / N
    rstd = 1 / tl.sqrt(var + eps)
    # Write mean / rstd
    tl.store(Mean + row, mean)
    tl.store(Rstd + row, rstd)
    # Normalize and apply linear transformation
    x_hat = (x - mean) * rstd

    # Scale calculation
    abs_x_hat = tl.abs(x_hat)
    max_val = tl.max(abs_x_hat, axis=0) + SCALE_MIN_THRES
    scale_output = max_val / fp8_max
    scale_output = scale_output.to(SY.type.element_ty)

    tl.store(SY + row, scale_output)

    # Write output
    tl.store(Y + cols, x_hat, mask=cols < N)


@triton.heuristics(
    {
        "BLOCK_SN": lambda args: args["N"] // args["QB"],
        "BLOCK_SN2": lambda args: args["N2"] // args["QB"],
    }
)
@triton.jit
def _layer_norm_bwd_dx_fused(
    DX,  # pointer to the input gradient
    DY,  # pointer to the output gradient
    X,  # pointer to the input
    SX,  # pointer to the input
    noise_ptr,  # noise for stochastic
    Mean,  # pointer to the mean
    Rstd,  # pointer to the 1/std
    stride,  # how much to increase the pointer when moving by 1 row
    scale_stride,  # how much to increase the pointer when moving by 1 row
    N: tl.constexpr,  # number of columns in X
    N2: tl.constexpr,  # number of columns in X
    SN: tl.constexpr,
    SN2: tl.constexpr,
    QB: tl.constexpr,
    SCALE_MIN_THRES: tl.constexpr,
    STOCHASTIC: tl.constexpr,
    BLOCK_SN: tl.constexpr,
    BLOCK_SN2: tl.constexpr,
):
    # Map the program id to the elements of X, DX, and DY it should compute.
    row = tl.program_id(0)
    cols = tl.arange(0, N2)
    scale_cols = tl.arange(0, SN2)
    mask = cols < N
    X += row * stride
    DY += row * stride
    DX += row * stride
    SX += row * scale_stride
    # Load data to SRAM
    x = tl.load(X + cols, mask=mask, other=0.0).to(tl.float32)
    scale_x = tl.load(SX + scale_cols, mask=scale_cols < SN, other=0.0).to(tl.float32)
    scale_x = tl.reshape(scale_x, (BLOCK_SN2, 1))
    x = tl.reshape(x, (BLOCK_SN2, QB))
    x = x * scale_x
    x = tl.reshape(x, N2)

    dy = tl.load(DY + cols, mask=mask, other=0.0).to(tl.float32)

    mean = tl.load(Mean + row)
    rstd = tl.load(Rstd + row)
    # Compute dx
    xhat = (x - mean) * rstd
    xhat = tl.where(mask, xhat, 0.0)
    dy = tl.where(mask, dy, 0.0)
    c1 = tl.sum(xhat * dy, axis=0) / N
    c2 = tl.sum(dy, axis=0) / N
    dx = (dy - (xhat * c1 + c2)) * rstd

    if STOCHASTIC:
        # noise_ptr += row * stride
        # noise_block_ptr = noise_ptr + cols
        # noise = tl.load(noise_block_ptr, mask=mask, other=0.)

        noise_offset = row * stride + cols
        noise = tl.rand(0, noise_offset)

        dx = _stochastic_rounding(dx, noise, e_bit, m_bit)

    dx = dx.to(DX.type.element_ty)

    # Write dx
    tl.store(DX + cols, dx, mask=mask)


def fp8_layernorm_noparam_forward(x, s_x, QB, eps, transpose_output_2d=False):
    # Change batched 3D input to 2D
    batched = False
    if len(x.shape) == 3:
        assert len(s_x.shape) == 3
        batched = True
        BS = x.shape[0]
        x = x.reshape(-1, x.shape[-1])
        s_x = s_x.reshape(-1, s_x.shape[-1])

    # allocate output
    M, N = x.shape
    _, SN = s_x.shape
    y = torch.empty_like(x, dtype=torch.float32)
    s_y = torch.empty(
        (M,), dtype=torch.bfloat16, device=x.device
    )  # We do this because we apply per-tensor quantization for it afterwards
    mean = torch.empty((M,), dtype=torch.float32, device=x.device)
    rstd = torch.empty((M,), dtype=torch.float32, device=x.device)
    # heuristics for number of warps
    num_warps = 8
    fp8MaxValue = FP8_MAX_VALUE[x.dtype]

    N2 = triton.next_power_of_2(N)
    SN2 = N2 // QB
    # enqueue kernel
    _layer_norm_fwd_fused[(M,)](  #
        x,
        s_x,
        y,
        s_y,
        mean,
        rstd,  #
        x.stride(0),
        s_x.stride(0),
        N,
        N2,
        SN,
        SN2,
        QB,
        eps,
        fp8MaxValue,
        SCALE_MIN_THRES=SCALE_MIN_THRES,  #
        num_warps=num_warps,
        num_ctas=1,
    )
    # reduction
    s_y_max = s_y.max()
    qy, s_y_max, qy_t = fp8_division_transpose(y, QB, x.dtype, s_y_max)

    # Recover 2D to 3D
    if batched:
        y = y.reshape(BS, -1, y.shape[-1])
        qy = qy.reshape(BS, -1, y.shape[-1])
        if not transpose_output_2d:
            qy_t = qy_t.reshape(BS, -1, y.shape[-1])

    return qy, s_y_max, qy_t, (mean, rstd, num_warps)


def fp8_layernorm_noparam_backward(x, s_x, g, QB, m, v, num_warps, stochastic=False):
    # Change batched 3D input to 2D
    batched = False
    if len(x.shape) == 3:
        assert len(s_x.shape) == 3
        batched = True
        BS = x.shape[0]
        x = x.reshape(-1, x.shape[-1])
        s_x = s_x.reshape(-1, s_x.shape[-1])
        g = g.reshape(-1, g.shape[-1])

    if stochastic:
        # noise = torch.empty_like(g, dtype=torch.float32).uniform_(-0.5, 0.5)
        noise = None
    else:
        noise = None

    # heuristics for amount of parallel reduction stream for DW/DB
    dx = torch.empty_like(g, dtype=torch.bfloat16)
    # enqueue kernel using forward pass heuristics
    # also compute partial sums for DW and DB
    M, N = g.shape
    _, SN = s_x.shape

    N2 = triton.next_power_of_2(N)
    SN2 = triton.next_power_of_2(SN)
    _layer_norm_bwd_dx_fused[(M,)](  #
        dx,
        g,
        x,
        s_x,
        noise,
        m,
        v,  #
        x.stride(0),
        s_x.stride(0),
        N,
        N2,
        SN,
        SN2,
        QB,
        SCALE_MIN_THRES=SCALE_MIN_THRES,
        STOCHASTIC=stochastic,
        num_warps=num_warps,
    )

    if batched:
        dx = dx.reshape(BS, -1, dx.shape[-1])

    return dx



# Copyright 2024 NVIDIA CORPORATION & AFFILIATES
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0

import torch


class FP8Manager:
    """Class to keep track of and manipulate the global
    FP8 state at different stages of execution.
    """

    is_first_microbatch = True



# Copyright 2024 NVIDIA CORPORATION & AFFILIATES
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0

import math
import sys
import unittest
from dataclasses import dataclass

import torch
import qoptim_cuda

from coat.activation.fake_quantization.quantize_function import (block_cut,
                                                                 block_quant,
                                                                 block_reshape)


@dataclass
class QArgs:
    row_blocksize: int = 1
    col_blocksize: int = 128
    pad_block: bool = True
    first_order_bit: str = "E4M3"
    second_order_bit: str = "E4M3"
    epsilon = 1e-8


qargs = QArgs()

# preparations
beta1, beta2 = 0.9, 0.95
lr, wd, eps = 4e-1, 0.1, 1e-8
step, qgroup_size = 1000, 128
RatioUpperbound = {
    "E4M3": 448 * 448 / 2,
}

M, N = 2, 128
torch.manual_seed(0)
torch.set_printoptions(sci_mode=False)
params = torch.randn((M, N), dtype=torch.float32, device="cuda")
grads = torch.randn((M, N), dtype=torch.float32, device="cuda")
params_py, params_cuda = params.clone(), params.clone()
grads_py, grads_cuda = grads.clone(), grads.clone()

fp_exp_avg = torch.randn((M, N), dtype=torch.float32, device="cuda")
fp_exp_avg_sq = torch.randn((M, N), dtype=torch.float32, device="cuda").abs()

power_options = torch.tensor([i / 16 for i in range(1, 129)]).to(fp_exp_avg)

""" Prepare the data """
# Quantize the first order momentum
Bexp_avg = block_cut(fp_exp_avg, qargs.row_blocksize, qargs.col_blocksize, pad_block=qargs.pad_block)
RQexp_avg, Qexp_avg, Sexp_avg = block_quant(
    Bexp_avg,
    True,
    qargs.first_order_bit,
    stochastic=False,
    epsilon=qargs.epsilon,
    apply_quantize=(qargs.first_order_bit != 100),
    layer_name="first_order",
)
Qexp_avg = block_reshape(Qexp_avg, fp_exp_avg, qargs.row_blocksize, qargs.col_blocksize, pad_block=qargs.pad_block)
fp_exp_avg = block_reshape(RQexp_avg, fp_exp_avg, qargs.row_blocksize, qargs.col_blocksize, pad_block=qargs.pad_block)

# Quantize the second order momentum
Bexp_avg_sq = block_cut(fp_exp_avg_sq, qargs.row_blocksize, qargs.col_blocksize, pad_block=qargs.pad_block)
RQexp_avg_sq, Qexp_avg_sq, Sexp_avg_sq = block_quant(
    Bexp_avg_sq,
    True,
    qargs.first_order_bit,
    stochastic=False,
    epsilon=qargs.epsilon,
    apply_quantize=(qargs.first_order_bit != 100),
    layer_name="first_order",
)
Qexp_avg_sq = block_reshape(
    Qexp_avg_sq, fp_exp_avg_sq, qargs.row_blocksize, qargs.col_blocksize, pad_block=qargs.pad_block
)
fp_exp_avg_sq = block_reshape(
    RQexp_avg_sq, fp_exp_avg_sq, qargs.row_blocksize, qargs.col_blocksize, pad_block=qargs.pad_block
)

Qexp_avg, Qexp_avg_sq = Qexp_avg.to(torch.float8_e4m3fn), Qexp_avg_sq.to(torch.float8_e4m3fn)
Sexp_avg, Sexp_avg_sq = Sexp_avg.to(torch.float32), Sexp_avg_sq.to(torch.float32)

# These are for coat's dynamic range expansion
IDexp_avg = torch.randint(power_options.size(0), size=Sexp_avg.shape)
IDexp_avg_sq = torch.randint(power_options.size(0), size=Sexp_avg_sq.shape)
EXPexp_avg = power_options[IDexp_avg].to(Sexp_avg).squeeze(2)
EXPexp_avg_sq = power_options[IDexp_avg_sq].to(Sexp_avg_sq).squeeze(2)
SQRTexp_avg = torch.randn_like(Sexp_avg).squeeze(2)
SQRTexp_avg_sq = torch.randn_like(Sexp_avg_sq).squeeze(2)

fp_exp_avg = fp_exp_avg.sign() * torch.pow(fp_exp_avg.abs(), 1 / EXPexp_avg) * SQRTexp_avg
fp_exp_avg_sq = torch.pow(fp_exp_avg_sq, 1 / EXPexp_avg_sq) * SQRTexp_avg_sq


def simulated_optimizer_expand_step(params, grads, exp_avg, exp_avg_sq, beta1, beta2, lr, wd, eps, step, qgroup_size):
    # update step
    step += 1

    # Perform stepweight decay
    params.mul_(1 - lr * wd)

    # Decay the first and second moment running average coefficient
    exp_avg.lerp_(grads, 1 - beta1)
    exp_avg_sq.mul_(beta2).addcmul_(grads, grads, value=1 - beta2)

    bias_correction1 = 1 - beta1**step
    bias_correction2 = 1 - beta2**step

    step_size = lr / bias_correction1

    bias_correction2_sqrt = math.sqrt(bias_correction2)

    denom = (exp_avg_sq.sqrt() / bias_correction2_sqrt).add_(eps)

    params.addcdiv_(fp_exp_avg, denom, value=-step_size)

    """ Quantize the first order momentum """
    fp_exp_shape = fp_exp_avg.shape
    q_exp_avg = fp_exp_avg.reshape(1, -1)

    Bexp_avg = block_cut(q_exp_avg, qargs.row_blocksize, qargs.col_blocksize, pad_block=qargs.pad_block)

    # Dynamic Range Expansion
    Bexp_avg_sign = Bexp_avg.sign()
    block_max, block_min = Bexp_avg.abs().max(dim=-1)[0], Bexp_avg.abs().min(dim=-1)[0]
    block_sqrt_minmax1 = (block_max * block_min).sqrt().unsqueeze(2)
    B1Ratio = block_max / block_min

    ratio_upperbound = RatioUpperbound[qargs.first_order_bit]
    B1ExpIdx = (torch.pow(B1Ratio, power_options) < ratio_upperbound).sum(dim=1)
    B1Exp = power_options[B1ExpIdx - 1].unsqueeze(1).unsqueeze(2)

    # re-center for numerical stability
    Bexp_avg = Bexp_avg / block_sqrt_minmax1
    Bexp_avg = torch.pow(Bexp_avg.abs(), B1Exp) * Bexp_avg_sign

    # Dequantize the output to FP32
    RQexp_avg, Qexp_avg, Sexp_avg = block_quant(
        Bexp_avg,
        True,
        qargs.first_order_bit,
        stochastic=False,
        epsilon=qargs.epsilon,
        apply_quantize=(qargs.first_order_bit != 100),
        layer_name="first_order",
    )
    Qexp_avg = block_reshape(Qexp_avg, q_exp_avg, qargs.row_blocksize, qargs.col_blocksize, pad_block=qargs.pad_block)

    RQexp_avg = torch.pow(RQexp_avg.abs(), 1 / B1Exp) * Bexp_avg_sign
    RQexp_avg = RQexp_avg * block_sqrt_minmax1

    RQexp_avg = block_reshape(RQexp_avg, q_exp_avg, qargs.row_blocksize, qargs.col_blocksize, pad_block=qargs.pad_block)
    rq_exp_avg = RQexp_avg.reshape(fp_exp_shape)

    fp_exp_avg.data = rq_exp_avg

    """ Quantize the second order momentum """
    fp_exp_shape = fp_exp_avg_sq.shape
    q_exp_avg_sq = fp_exp_avg_sq.reshape(1, -1)
    Bexp_avg_sq = block_cut(q_exp_avg_sq, qargs.row_blocksize, qargs.col_blocksize, pad_block=qargs.pad_block)

    # Dynamic Range Expansion
    block_max, block_min = Bexp_avg_sq.max(dim=-1)[0], Bexp_avg_sq.min(dim=-1)[0]
    block_sqrt_minmax2 = (block_max * block_min).sqrt().unsqueeze(2)
    B2Ratio = block_max / block_min

    ratio_upperbound = RatioUpperbound[qargs.second_order_bit]
    B2ExpIdx = (torch.pow(B2Ratio, power_options) < ratio_upperbound).sum(dim=1)
    B2Exp = power_options[B2ExpIdx - 1].unsqueeze(1).unsqueeze(2)

    # re-center for numerical stability
    Bexp_avg_sq = Bexp_avg_sq / block_sqrt_minmax2
    Bexp_avg_sq = torch.pow(Bexp_avg_sq, B2Exp)

    # Dequantize the output to FP32
    RQexp_avg_sq, Qexp_avg_sq, Sexp_avg_sq = block_quant(
        Bexp_avg_sq,
        True,
        qargs.second_order_bit,
        stochastic=False,
        epsilon=qargs.epsilon,
        apply_quantize=(qargs.second_order_bit != 100),
        layer_name="second_order",
    )
    Qexp_avg_sq = block_reshape(
        Qexp_avg_sq, q_exp_avg_sq, qargs.row_blocksize, qargs.col_blocksize, pad_block=qargs.pad_block
    )

    RQexp_avg_sq = torch.pow(RQexp_avg_sq.abs(), 1 / B2Exp)
    RQexp_avg_sq = RQexp_avg_sq * block_sqrt_minmax2

    RQexp_avg_sq = block_reshape(
        RQexp_avg_sq, q_exp_avg_sq, qargs.row_blocksize, qargs.col_blocksize, pad_block=qargs.pad_block
    )
    rq_exp_avg_sq = RQexp_avg_sq.reshape(fp_exp_shape)

    fp_exp_avg_sq.data = rq_exp_avg_sq

    return Qexp_avg, Sexp_avg, B1Exp, block_sqrt_minmax1, Qexp_avg_sq, Sexp_avg_sq, B2Exp, block_sqrt_minmax2


def cuda_optimizer_expand_step(
    params,
    grads,
    exp_avg,
    scale_exp_avg,
    expand_exp_avg,
    sqrt_minmax_exp_avg,
    exp_avg_sq,
    scale_exp_avg_sq,
    expand_exp_avg_sq,
    sqrt_minmax_exp_avg_sq,
    beta1,
    beta2,
    lr,
    wd,
    eps,
    step,
    qgroup_size,
    expand_min,
):
    qoptim_cuda.fp8_adamw_expand_step(
        params,
        grads,
        exp_avg,
        scale_exp_avg,
        expand_exp_avg,
        sqrt_minmax_exp_avg,
        exp_avg_sq,
        scale_exp_avg_sq,
        expand_exp_avg_sq,
        sqrt_minmax_exp_avg_sq,
        beta1,
        beta2,
        lr,
        wd,
        eps,
        step,
        qgroup_size,
        expand_min,
    )


class TestExpandAdamW(unittest.TestCase):
    def test_expand_coatadamw(self):
        (
            Qexp_avg_py,
            Sexp_avg_py,
            B1Exp,
            block_sqrt_minmax1,
            Qexp_avg_sq_py,
            Sexp_avg_sq_py,
            B2Exp,
            block_sqrt_minmax2,
        ) = simulated_optimizer_expand_step(
            params_py, grads_py, fp_exp_avg, fp_exp_avg_sq, beta1, beta2, lr, wd, eps, step, qgroup_size
        )

        cuda_optimizer_expand_step(
            params_cuda,
            grads_cuda,
            Qexp_avg,
            Sexp_avg,
            EXPexp_avg,
            SQRTexp_avg,
            Qexp_avg_sq,
            Sexp_avg_sq,
            EXPexp_avg_sq,
            SQRTexp_avg_sq,
            beta1,
            beta2,
            lr,
            wd,
            eps,
            step,
            qgroup_size,
            16,
        )

        self.assertTrue(torch.allclose(params_py.flatten(), params_cuda.flatten(), 1e-4))
        self.assertTrue(
            torch.allclose(Qexp_avg_py.to(torch.bfloat16).flatten(), Qexp_avg.to(torch.bfloat16).flatten(), 1e-4)
        )
        self.assertTrue(torch.allclose(Sexp_avg_py.flatten(), Sexp_avg.flatten(), 1e-4))
        self.assertTrue(torch.allclose(B1Exp.flatten(), EXPexp_avg.flatten(), 1e-4))
        self.assertTrue(torch.allclose(block_sqrt_minmax1.flatten(), SQRTexp_avg.flatten(), 1e-4))
        self.assertTrue(
            torch.allclose(Qexp_avg_sq_py.to(torch.bfloat16).flatten(), Qexp_avg_sq.to(torch.bfloat16).flatten(), 1e-4)
        )
        self.assertTrue(torch.allclose(Sexp_avg_sq_py.flatten(), Sexp_avg_sq.flatten(), 1e-4))
        self.assertTrue(torch.allclose(B2Exp.flatten(), EXPexp_avg_sq.flatten(), 1e-4))
        self.assertTrue(torch.allclose(block_sqrt_minmax2.flatten(), SQRTexp_avg_sq.flatten(), 1e-4))


if __name__ == "__main__":
    torch.manual_seed(0)
    torch.set_printoptions(sci_mode=False, linewidth=200)
    unittest.main()



# Copyright 2024 NVIDIA CORPORATION & AFFILIATES
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0

import torch
# 4 block
import triton
import triton.language as tl
from triton.language.extra.cuda import libdevice

from ._division_transpose import fp8_division_transpose
from .common import (FP8_MAX_VALUE, SCALE_MIN_THRES, convert_fp8_to_embit,
                     convert_str_to_fp8, get_configs_io_block)

"""Element-wise Multiplication Backward"""
"""Input1 (Gate) uses 1 * 16 group quantization"""
"""Input2 (Up) uses 1 * 16 group quantization"""
"""Grad (Down) uses full-precision/BF16"""
"""Output1 (Gate) uses full-precision/BF16"""
"""Output2 (Up) uses per-tensor quantization, we can choose whether it should be quantized inside this function"""
"""The input can be 2D or 3D, but the calculation is performed in 2D"""


@triton.autotune(
    configs=[] + get_configs_io_block(),
    key=[
        "M",
        "N",
    ],
)
@triton.heuristics(
    {
        "BLOCK_SN": lambda args: args["BLOCK_N"] // args["QB"],
    }
)
@triton.jit
def _fp8_mul_backward_kernel(
    output1_ptr,  # output
    output2_ptr,
    output2_scale_ptr,  # output
    input1_ptr,
    input1_scale_ptr,  # input
    input2_ptr,
    input2_scale_ptr,  # input
    grad_ptr,  # input
    M,
    N,
    SN,
    QB: tl.constexpr,
    fp8_max,
    e_bit,
    m_bit,  # shape
    input1_stride_0,
    input1_stride_1,  # input1 stride
    s_input1_stride_0,
    s_input1_stride_1,  # scale of input1 stride
    input2_stride_0,
    input2_stride_1,  # input2 stride
    s_input2_stride_0,
    s_input2_stride_1,  # scale of input2 stride
    grad_stride_0,
    grad_stride_1,  # input stride
    output1_stride_0,
    output1_stride_1,  # output stride
    output2_stride_0,
    output2_stride_1,  # output stride
    s_output2_stride_0,
    s_output2_stride_1,  # scale of output stride
    SCALE_MIN_THRES: tl.constexpr,
    STOCHASTIC: tl.constexpr,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
    BLOCK_SN: tl.constexpr,
):  # CUDA block size

    # Block PID
    pid = tl.program_id(0)
    NUM_BLOCK_N = tl.cdiv(N, BLOCK_N)
    pid_dim0 = pid // NUM_BLOCK_N
    pid_dim1 = pid % NUM_BLOCK_N

    # --- The first input ---
    input1_block_ptr = tl.make_block_ptr(
        base=input1_ptr,
        shape=(M, N),
        strides=(input1_stride_0, input1_stride_1),
        offsets=(pid_dim0 * BLOCK_M, pid_dim1 * BLOCK_N),
        block_shape=(BLOCK_M, BLOCK_N),
        order=(1, 0),
    )

    # input ptr
    scale_input1_ptr = tl.make_block_ptr(
        base=input1_scale_ptr,
        shape=(M, SN),
        strides=(s_input1_stride_0, s_input1_stride_1),
        offsets=(pid_dim0 * BLOCK_M, pid_dim1 * BLOCK_SN),
        block_shape=(BLOCK_M, BLOCK_SN),
        order=(1, 0),
    )

    input1 = tl.load(input1_block_ptr)
    scale_input1 = tl.load(scale_input1_ptr)

    input1 = input1.to(tl.float32)
    scale_input1 = scale_input1.to(tl.float32)

    # Dequantize and mul calculation
    scale_input1 = tl.reshape(scale_input1, (BLOCK_M, BLOCK_SN, 1))
    input1 = tl.reshape(input1, (BLOCK_M, BLOCK_SN, QB))
    input1 = input1 * scale_input1

    # --- The second input ---
    input2_block_ptr = tl.make_block_ptr(
        base=input2_ptr,
        shape=(M, N),
        strides=(input2_stride_0, input2_stride_1),
        offsets=(pid_dim0 * BLOCK_M, pid_dim1 * BLOCK_N),
        block_shape=(BLOCK_M, BLOCK_N),
        order=(1, 0),
    )

    # input ptr
    scale_input2_ptr = tl.make_block_ptr(
        base=input2_scale_ptr,
        shape=(M, SN),
        strides=(s_input2_stride_0, s_input2_stride_1),
        offsets=(pid_dim0 * BLOCK_M, pid_dim1 * BLOCK_SN),
        block_shape=(BLOCK_M, BLOCK_SN),
        order=(1, 0),
    )

    input2 = tl.load(input2_block_ptr)
    scale_input2 = tl.load(scale_input2_ptr)

    input2 = input2.to(tl.float32)
    scale_input2 = scale_input2.to(tl.float32)

    # Dequantize and mul calculation
    scale_input2 = tl.reshape(scale_input2, (BLOCK_M, BLOCK_SN, 1))
    input2 = tl.reshape(input2, (BLOCK_M, BLOCK_SN, QB))
    input2 = input2 * scale_input2

    # pointers of gradient
    grad_block_ptr = tl.make_block_ptr(
        base=grad_ptr,
        shape=(M, N),
        strides=(grad_stride_0, grad_stride_1),
        offsets=(pid_dim0 * BLOCK_M, pid_dim1 * BLOCK_N),
        block_shape=(BLOCK_M, BLOCK_N),
        order=(1, 0),
    )

    grad = tl.load(grad_block_ptr)
    grad = grad.to(tl.float32)
    grad = tl.reshape(grad, (BLOCK_M, BLOCK_SN, QB))

    # Actual Calculation of Mul Backward
    grad1 = grad * input2
    grad1 = tl.reshape(grad1, (BLOCK_M, BLOCK_N))
    grad1 = grad1.to(output1_ptr.type.element_ty)

    # pointers
    output1_block_ptr = tl.make_block_ptr(
        base=output1_ptr,
        shape=(M, N),
        strides=(output1_stride_0, output1_stride_1),
        offsets=(pid_dim0 * BLOCK_M, pid_dim1 * BLOCK_N),
        block_shape=(BLOCK_M, BLOCK_N),
        order=(1, 0),
    )
    tl.store(output1_block_ptr, grad1)

    # Actual Calculation of Mul Backward
    grad2 = grad * input1
    # Quantize the grad 1 - Scale calculation
    abs_grad2 = tl.abs(grad2)
    max_val = tl.max(abs_grad2, axis=2) + SCALE_MIN_THRES
    scale_grad2 = max_val / fp8_max
    scale_grad2 = tl.reshape(scale_grad2, (BLOCK_M, BLOCK_SN, 1))
    # Quantize
    # grad1 = tl.fdiv(grad1, scale_output) # do not quantize the output due to the data flow
    grad2 = grad2.to(output2_ptr.type.element_ty)
    scale_grad2 = scale_grad2.to(output2_scale_ptr.type.element_ty)
    scale_grad2 = tl.reshape(scale_grad2, (BLOCK_M, BLOCK_SN))
    grad2 = tl.reshape(grad2, (BLOCK_M, BLOCK_N))

    # pointers
    output2_block_ptr = tl.make_block_ptr(
        base=output2_ptr,
        shape=(M, N),
        strides=(output2_stride_0, output2_stride_1),
        offsets=(pid_dim0 * BLOCK_M, pid_dim1 * BLOCK_N),
        block_shape=(BLOCK_M, BLOCK_N),
        order=(1, 0),
    )
    scale_output2_ptr = tl.make_block_ptr(
        base=output2_scale_ptr,
        shape=(M, SN),
        strides=(s_output2_stride_0, s_output2_stride_1),
        offsets=(pid_dim0 * BLOCK_M, pid_dim1 * BLOCK_SN),
        block_shape=(BLOCK_M, BLOCK_SN),
        order=(1, 0),
    )
    tl.store(output2_block_ptr, grad2)
    tl.store(scale_output2_ptr, scale_grad2)


def fp8_mul_backward(
    x1, s_x1, x2, s_x2, g, QB, fp8type, stochastic=False, output_quantized_transpose=False
):  # Stochastic Rounding is left outside this function
    # Change batched 3D input to 2D
    batched = False
    if len(x1.shape) == 3:
        assert len(s_x1.shape) == 3
        batched = True
        BS = x1.shape[0]
        x1 = x1.reshape(-1, x1.shape[-1])
        s_x1 = s_x1.reshape(-1, s_x1.shape[-1])
        x2 = x2.reshape(-1, x2.shape[-1])
        s_x2 = s_x2.reshape(-1, s_x2.shape[-1])
        g = g.reshape(-1, g.shape[-1])

    if stochastic:
        noise = torch.empty_like(g, dtype=torch.float32).uniform_(-0.5, 0.5)
    else:
        noise = None

    # defining the input and output tensor
    M, N = x1.shape
    _, SN = s_x1.shape  # assume the shape of quantization block size is always 1 * G
    assert x1.shape == x2.shape
    assert s_x1.shape == s_x2.shape

    if isinstance(fp8type, str):
        fp8type = convert_str_to_fp8[fp8type]
    y1 = torch.empty_like(g, dtype=torch.bfloat16)
    y2 = torch.empty_like(g, dtype=torch.bfloat16)
    s_y2 = torch.empty_like(s_x1, dtype=torch.bfloat16)
    fp8MaxValue = FP8_MAX_VALUE[fp8type]  # E4M3 and E5M2 have different max value
    e_bit, m_bit = convert_fp8_to_embit[fp8type]

    grid = lambda META: (triton.cdiv(M, META["BLOCK_M"]) * triton.cdiv(N, META["BLOCK_N"]),)

    _fp8_mul_backward_kernel[grid](
        y1,
        y2,
        s_y2,
        x1,
        s_x1,
        x2,
        s_x2,
        g,
        M,
        N,
        SN,
        QB,
        fp8MaxValue,
        e_bit,
        m_bit,
        x1.stride(0),
        x1.stride(1),
        s_x1.stride(0),
        s_x1.stride(1),
        x2.stride(0),
        x2.stride(1),
        s_x2.stride(0),
        s_x2.stride(1),
        g.stride(0),
        g.stride(1),
        y1.stride(0),
        y1.stride(1),
        y2.stride(0),
        y2.stride(1),
        s_y2.stride(0),
        s_y2.stride(1),
        SCALE_MIN_THRES=SCALE_MIN_THRES,
        STOCHASTIC=stochastic,
    )

    if not output_quantized_transpose:
        # Recover 2D to 3D
        if batched:
            y1 = y1.reshape(BS, -1, y1.shape[-1])

            y2 = y2.reshape(BS, -1, y2.shape[-1])
            s_y2 = s_y2.reshape(BS, -1, s_y2.shape[-1])

        return y1, (y2, s_y2)
    else:
        # Per-tensor quantization
        s_y2_max = s_y2.max()
        qy2, s_y2_max, qy2_t = fp8_division_transpose(y2, QB, fp8type, s_y2_max)

        # Recover 2D to 3D
        if batched:
            y1 = y1.reshape(BS, -1, y1.shape[-1])

            y2 = y2.reshape(BS, -1, y2.shape[-1])
            qy2 = qy2.reshape(BS, -1, qy2.shape[-1])
            s_y2 = s_y2.reshape(BS, -1, s_y2.shape[-1])

        return y1, (qy2, s_y2_max, qy2_t)



# Copyright 2024 NVIDIA CORPORATION & AFFILIATES
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0

from setuptools import setup
from torch.utils.cpp_extension import BuildExtension, CUDAExtension

setup(
    name="qoptim_cuda",
    ext_modules=[
        CUDAExtension(
            name="qoptim_cuda",
            sources=[
                "fp8_adamw_cuda.cpp",
                "fp8_adamw_cuda_kernel.cu",
                "fp8_adamw_expand_cuda.cpp",
                "fp8_adamw_expand_cuda_kernel.cu",
                "bindings.cpp",
            ],
            # include_dirs=[
            #     'include'
            # ],
            extra_compile_args={
                "nvcc": [
                    "-O3",
                    "-std=c++17",
                    "-DTORCH_USE_CUDA_DSA",
                    "-U__CUDA_NO_HALF_OPERATORS__",
                    "-U__CUDA_NO_HALF_CONVERSIONS__",
                    "-U__CUDA_NO_BFLOAT16_CONVERSIONS__",
                    "-U__CUDA_NO_HALF2_OPERATORS__",
                ]
            },
        ),
    ],
    cmdclass={"build_ext": BuildExtension},
)



# Copyright 2024 NVIDIA CORPORATION & AFFILIATES
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0

import unittest

import torch

from coat.activation.real_quantization.common import FP8_MAX_VALUE


def random_tensor_generator(BS, SL, CDIM, QB, fp8type):
    """
    Generates a random tensor, quantizes it, and returns the quantized result.

    Parameters:
    - BS (int): Batch size.
    - SL (int): Sequence length.
    - CDIM (int): Channel dimension (input dimension).
    - QB (int): Quantization block size. Should be the power of 2.
    - fp8type (torch.dtype): Target data type for quantization (e.g., torch.float8_e4m3fn, torch.float8_e5m2).

    Returns:
    - rqx (torch.Tensor): Reconstructed tensor after quantization. Should be equal to qx * sx
    - qx (torch.Tensor): Quantized tensor.
    - sx (torch.Tensor): Scaling factors used in quantization.
    """
    x = torch.randn(BS, SL, CDIM).cuda()
    rqx, qx, sx = quantize_tensor(x, BS, SL, CDIM, QB, fp8type)
    return rqx, qx, sx


def quantize_tensor(x, BS, SL, CDIM, QB, fp8type, quant_type="per_group"):
    """
    Quantizes a given tensor using a per-group quantization strategy, converting it into fp8 format.
    """
    if quant_type == "per_tensor":
        sx = (x.abs().max() / FP8_MAX_VALUE[fp8type]).to(torch.bfloat16)
        qx = (x / sx).to(fp8type)
        rqx = qx.to(torch.float32) * sx.to(torch.float32)
    elif quant_type == "per_block":
        _qx = x.reshape(BS * SL // QB, QB, CDIM // QB, QB)
        _qx = _qx.permute(0, 2, 1, 3)
        sx = _qx.abs().amax(dim=(2,3)) / FP8_MAX_VALUE[fp8type]
        sx = sx.to(torch.bfloat16)
        _qx = (_qx / sx.unsqueeze(2).unsqueeze(3)).to(fp8type)
        qx = _qx.permute(0, 2, 1, 3).reshape(BS * SL, CDIM)
        rqx = (_qx.float() * sx.unsqueeze(2).unsqueeze(3).float()).permute(0, 2, 1, 3).reshape(BS * SL, CDIM)
    elif quant_type == "per_group":
        _qx = x.reshape(BS, SL, CDIM // QB, QB)
        sx = _qx.abs().amax(dim=(3)) / FP8_MAX_VALUE[fp8type]
        sx = sx.to(torch.bfloat16)
        _qx = (_qx / sx.unsqueeze(3)).to(fp8type)
        qx = _qx.reshape(BS, SL, CDIM)
        rqx = (_qx.float() * sx.unsqueeze(3).float()).reshape(BS, SL, CDIM)
    else:
        raise ValueError(f"Invalid quantization type {quant_type}. ")
    return rqx, qx, sx

def quantize_perblock(x, BS, SL, CDIM, QB, fp8type):
    """
    Quantizes a given tensor using a per-block quantization strategy, converting it into fp8 format.
    """
    sx = (x.abs().max() / FP8_MAX_VALUE[fp8type]).to(torch.bfloat16)
    qx = (x / sx).to(fp8type)
    rqx = qx.to(torch.float32) * sx.to(torch.float32)
    return rqx, qx, sx

def dequantize_tensor(x_triton, s_triton, BS, SL, CDIM, QB):
    """
    Dequantizes a tensor by converting it back to full precision using scaling factors.
    """
    if s_triton.numel() == 1:
        # In this case it should be per-tensor quantization
        output_triton = x_triton.to(torch.float32) * s_triton.to(torch.float32)
    else:
        # In this case it should be per-group quantization with group size 1 * G
        assert len(s_triton.shape) == 3
        _x_triton = x_triton.reshape(BS, SL, CDIM // QB, QB)
        _x_triton = _x_triton.to(torch.float32)
        s_triton = s_triton.unsqueeze(3).to(torch.float32)
        output_triton = (_x_triton * s_triton).reshape(BS, SL, CDIM)
    return output_triton


def check_similarity(output_torch, output_triton, threshold=0.95):
    """
    Check if at least a specified percentage (default 95%) of values in two tensors are exactly the same.

    Parameters:
    - output_torch (torch.Tensor): First tensor to compare.
    - output_triton (torch.Tensor): Second tensor to compare.
    - threshold (float): The percentage of matching elements required (default is 99%).

    Returns:
    - bool: True if the percentage of matching elements is above the threshold, False otherwise.
    """
    # Compare the two tensors element-wise for equality
    equal_elements = torch.eq(output_torch, output_triton)

    # Calculate the ratio of equal elements
    num_equal_elements = torch.sum(equal_elements).item()
    total_elements = output_torch.numel()
    similarity_ratio = num_equal_elements / total_elements

    # Check if the similarity ratio meets the threshold
    return similarity_ratio >= threshold



# Copyright 2024 NVIDIA CORPORATION & AFFILIATES
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0

import torch
import triton

SCALE_MIN_THRES = 1e-10

FP8_MAX_VALUE = {
    torch.float8_e4m3fn: 448.0,
    torch.float8_e5m2: 57344.0,
}

convert_str_to_fp8 = {"E4M3": torch.float8_e4m3fn, "E5M2": torch.float8_e5m2}
convert_fp8_to_embit = {
    torch.float8_e4m3fn: (4.0, 3.0),
    torch.float8_e5m2: (5.0, 2.0),
}


def get_configs_io_block(QB=16):
    configs = []
    for nstages in [3, 4]:
        for block_m in [QB, 2 * QB]:
            for block_n in [QB, QB * 2, 4 * QB]:
                for nwarps in [4, 8]:
                    configs.append(
                        triton.Config(
                            {"BLOCK_M": block_m, "BLOCK_N": block_n},
                            num_stages=nstages,
                            num_warps=nwarps,
                        )
                    )
    return configs

def get_tuned_kernel(kernel, QB=16):
    configs = get_configs_io_block(QB)
    return triton.autotune(configs=configs, key=["M", "N"])(kernel)

# from .common import SCALE_MIN_THRES, FP8_MAX_VALUE
#                     SCALE_MIN_THRES: tl.constexpr,
#  + SCALE_MIN_THRES
# SCALE_MIN_THRES=SCALE_MIN_THRES,



# Copyright 2024 NVIDIA CORPORATION & AFFILIATES
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0

import torch
# 4 block
import triton
import triton.language as tl
from triton.language.extra.cuda import libdevice

from .common import (FP8_MAX_VALUE, SCALE_MIN_THRES, convert_fp8_to_embit,
                     convert_str_to_fp8, get_configs_io_block)

"""Quantize Operator"""
"""Input uses full precision"""
"""Output uses 128 * 128 group quantization"""
"""The input can be 2D or 3D, but the calculation is performed in 2D"""


@triton.jit
def _fp8_quantize_perblock_kernel(
    output_ptr,
    output_scale_ptr,  # output
    input_ptr,  # input
    M,
    N,
    SM,
    SN,
    QB: tl.constexpr,
    fp8_max,  # shape
    input_stride_0,
    input_stride_1,  # input stride
    output_stride_0,
    output_stride_1,  # output stride
    s_output_stride_0,
    s_output_stride_1,  # scale of output stride
    SCALE_MIN_THRES: tl.constexpr,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
):  # CUDA block size

    # Block PID
    pid = tl.program_id(0)
    NUM_BLOCK_N = tl.cdiv(N, BLOCK_N)
    pid_dim0 = pid // NUM_BLOCK_N
    pid_dim1 = pid % NUM_BLOCK_N

    # pointers
    input_block_ptr = tl.make_block_ptr(
        base=input_ptr,
        shape=(M, N),
        strides=(input_stride_0, input_stride_1),
        offsets=(pid_dim0 * BLOCK_M, pid_dim1 * BLOCK_N),
        block_shape=(BLOCK_M, BLOCK_N),
        order=(1, 0),
    )

    input = tl.load(input_block_ptr)
    output = input.to(tl.float32)

    # Quantize Scale calculation
    abs_output = tl.abs(output)
    max_val = tl.max(abs_output) + SCALE_MIN_THRES
    scale_output = max_val / fp8_max

    # Quantize
    output = tl.fdiv(output, scale_output)

    output = output.to(output_ptr.type.element_ty)

    scale_output = scale_output.to(output_scale_ptr.type.element_ty)

    # pointers
    output_block_ptr = tl.make_block_ptr(
        base=output_ptr,
        shape=(M, N),
        strides=(output_stride_0, output_stride_1),
        offsets=(pid_dim0 * BLOCK_M, pid_dim1 * BLOCK_N),
        block_shape=(BLOCK_M, BLOCK_N),
        order=(1, 0),
    )
    scale_output_ptr = tl.make_block_ptr(
        base=output_scale_ptr,
        shape=(SM, SN),
        strides=(s_output_stride_0, s_output_stride_1),
        offsets=(pid_dim0, pid_dim1),
        block_shape=(1, 1),
        order=(1, 0),
    )

    tl.store(output_block_ptr, output)
    tl.store(scale_output_ptr, scale_output)


def fp8_quantize_perblock(x, QB, fp8type, scale_dtype=torch.bfloat16):
    # Change batched 3D input to 2D
    batched = False
    if len(x.shape) == 3:
        batched = True
        BS = x.shape[0]
        x = x.reshape(-1, x.shape[-1])

    # defining the input and output tensor
    M, N = x.shape
    SM, SN = M // QB, N // QB

    if isinstance(fp8type, str):
        fp8type = convert_str_to_fp8[fp8type]
    y = torch.empty_like(x, dtype=fp8type)
    s_y = torch.empty((SM, SN), dtype=scale_dtype, device=x.device)
    fp8MaxValue = FP8_MAX_VALUE[fp8type]  # E4M3 and E5M2 have different max value

    grid = lambda META: (triton.cdiv(M, META["BLOCK_M"]) * triton.cdiv(N, META["BLOCK_N"]),)

    _fp8_quantize_perblock_kernel[grid](
        y,
        s_y,
        x,
        M,
        N,
        SM,
        SN,
        QB,
        fp8MaxValue,
        x.stride(0),
        x.stride(1),
        y.stride(0),
        y.stride(1),
        s_y.stride(0),
        s_y.stride(1),
        SCALE_MIN_THRES=SCALE_MIN_THRES,
        BLOCK_M=QB,
        BLOCK_N=QB
    )

    # Do not recover to 3D

    return y, s_y
